#!/usr/bin/env python
# -*- coding: utf8 -*-

#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict, Counter
from datetime import datetime

from fmqlutils import VISTA_DATA_BASE_DIR
from fmqlutils.cacher.cacherUtils import SCHEMA_LOCN_TEMPL 
from fmqlutils.reporter.reportUtils import MarkdownTable, reportPercent, reportAbsAndPercent
from fmqlreports.webReportUtils import SITE_DIR_TEMPL, TOP_MD_TEMPL, keyStats, roundFloat, vistasOfVISNByOne

from fmqlreports.webReportUtils import muPlotRef, makePlots 
from patientDirectory import CommonPatientDirectory
from userDirectory import CommonUserDirectory

from webReportUserDirectory import csvDatesBySNO
        
"""
See note in user directory (as //)

... TODO: will split more as dive in (patient's common far less + need to make
up for lack of patient dates ... 
"""

# ###################################### Totals  ############################

PLOT_PREFIX = "visn20Patient"

def makeTotalsSection(cd, vistaNames, imageDir="Images"):
                       
    plotData = {}
                       
    mu = "## Patient Breakdown\n\n"
               
    countsFromTotal = { 
        "__entityName": "Patients",
    
        "__total": cd.total(),
        
        "> 1 VistA": cd.gt1VistACount(), # SHOULD BE IN OWN GRAPHIC
            
        "Male": cd.males(),
        "Marked Deceased": cd.deceased(),
        "Users Too": cd.usersToo(), 
    
        "FICN": cd.idPresence(CommonPatientDirectory.ENTRY_FICN_IDX), 
        "SSN": cd.idPresence(CommonPatientDirectory.ENTRY_SSN_IDX),
        "SCN": cd.idPresence(CommonPatientDirectory.ENTRY_SCN_IDX),
        "No ID": cd.noIds()
    }
    
    plotData["patientPercs"] = {
            "title": f'{countsFromTotal["__total"]} Patients',
            "plotName": f'{PLOT_PREFIX}Percs',
            "plotMethod": "plotIsHasBH",
            "specs": countsFromTotal
    }
    blurb = "The directory holds <span class='yellowIt'>{:,}</span> patients. <span class='yellowIt'>{}</span> are Male and nearly all have Integration Control Numbers (ICN) and Social Security Numbers (SSN). <span class='yellowIt'>{}</span> are marked deceased in at least one VistA holding part of their medical records. Note that <span class='yellowIt'>{}</span> of patients are also VistA users.".format(
        countsFromTotal["__total"],
        reportPercent(countsFromTotal["Male"], countsFromTotal["__total"]),
        reportPercent(countsFromTotal["Marked Deceased"], countsFromTotal["__total"]),
        reportPercent(countsFromTotal["Users Too"], countsFromTotal["__total"])
    )
    mu += "{}\n\n{}\n\n".format(
        blurb, 
        muPlotRef(plotData["patientPercs"], imageDir)
    )
        
    commonTotals = defaultdict(dict) # not picturing           
    allCnt = len(cd.vistas())
    for vistaInfo in cd.vistas():
        stationNo = vistaInfo["stationNumber"]
        stationMN = vistaNames[stationNo]
        commonCountsNos = cd.commonCounts(stationNo)
        commonCountsPer = {
            "all": commonCountsNos[allCnt],
            "some": sum(commonCountsNos[i] for i in range(2, allCnt)),
            "exclusive": commonCountsNos[1]
        }
        commonTotals[stationMN] = commonCountsPer
        
    plotData["commonByVistA"] = {
            "title": "Common Patient Count",
            "plotName": f'{PLOT_PREFIX}Common',
            "plotMethod": "plotCategoryBSV",
            "specs": commonTotals
    }
    blurb = "The graph below shows the total number of patients in each system as well as a breakdown of whether those patients are unique to that system (exclusive), in some but not every other system (some) or in every other (all) systems in the directory. Unlike the equivalent graph for user directory, this seems to show that systems big and small have similar proportions of unique and shared patients, but they don't."
    mu += "{}\n\n{}\n\n".format(
        blurb,
        muPlotRef(plotData["commonByVistA"], imageDir)
    )
    
    plotData["commonByVistAPerc"] = {
            "title": "Common Patient Count (Percentage)",
            "plotName": f'{PLOT_PREFIX}PercCommon',
            "plotMethod": "plotCategoryBSV",
            "specs": commonTotals,
            "kargs": { 
                "usePerc": True
            }
    }
    blurb = "The percentage view makes it clearer that smaller systems such as Walla Walla (WWW) through Roseland (ROS) have proportionately far more shared patients than their larger peers such as Puget Sound (PUG) or Portland (POR). This may be due to interfacility-consult work where a larger facility performs care for the patients of a smaller peer or because patients must go to bigger institutions for specialized care."
    mu += "{}\n\n{}\n\n".format(
        blurb,
        muPlotRef(plotData["commonByVistAPerc"], imageDir)
    )
        
    return plotData, mu
    
# #################################### Time Series Section ##########################

"""
Stories shown Patient:
- last two full months the patient addition is as expected, weekend and holiday drops and Puget is biggest, Walla Walla is smallest. It does show that unlike other systems, Puget does add patients, albeit at a lower number than usual, on holidays such as 1/20 and 2/17.
- pull back to yearly trends and again Puget is biggest but its gap on Portland only begins to grow in the last ten years
- monthly resamples reflect just what's expected. 
- addition very even in last five years in order expected
"""

def makeTimeSeriesSection(cd, vistaNames, imageDir="Images"):

    plotData = {}
    end = "2020-02"

    mu = "## Patient Additions over Time\n\n"
    
    dailiesDFCSV = csvDatesBySNO(cd)
            
    plotData["tsDailies"] = {
        "title": "2020/01-2020/02 Patient Addition",
        "plotName": f'{PLOT_PREFIX}',
        "plotMethod": "plotDailies",
        "dfcsv": dailiesDFCSV,
        "kargs": {
            "colNameById": vistaNames,
            "start": "2020-01",
            "end": end
        }
    }
    blurb = "The following shows __Daily Patient Addition__ for the last full three months for which data is available with Mondays marked on the X-axis. As expected addition \"dances\" up on weekdays and down on weekends which show little activity. For every system but Puget (PUG), holidays (1/20, 2/17) act like weekends. In contrast with the _User Directory_, this directory's additions mirror system size - Puget exceeds Portland (POR), all the way down the smallest, Walla Walla (WWW)."
    mu += "{}\n\n{}\n\n".format(
        blurb, 
        muPlotRef(
            plotData["tsDailies"],
            imageDir
        )
    )
        
    plotData["ts365RollingMean"] = {
        "title": "Trends in Patient Addition (365 day rolling)",
        "plotName": f'{PLOT_PREFIX}',
        "plotMethod": "plot365RollingMean",
        "dfcsv": dailiesDFCSV,
        "kargs": {
            "colNameById": vistaNames,
            "start": "1987-01", 
            "end": end
        }
    }
    blurb = "Here are the yearly trends using a __\"365 day rolling mean\"__ starting when the directory first gets data to the latest additions. Again, additions follow system size but first in 2000 and then in 2009, Puget and Portland jump up in volume compared to their  smaller peers with Puget's magnitude of increase exceeding Portland's. Except for Roseland, which seems to follow its own way. The reason for this - new types of care or additional clinics supported in systems? - should be investigated."
    mu += "{}\n\n{}\n\n".format(
        blurb,
        muPlotRef(
            plotData["ts365RollingMean"],
            imageDir
        )
    )
    
    plotData["tsMonthlyResamples"] = {
        "title": "Patients - Resampled Monthly (2018-{})".format(re.sub(r'\-', '/', end)),
        "plotName": f'{PLOT_PREFIX}',
        "plotMethod": "plotMonthlyResamples",
        "dfcsv": dailiesDFCSV,
        "kargs": {
            "colNameById": vistaNames,
            "start": '2018-01', 
            "end": end
        }
    }
    blurb = "A __monthly resample__ from 2018 to the latest common date echo's the yearly trends - unlike for User Directory, there are no unpredicted peaks or troughs."
    mu += "{}\n\n{}\n\n".format(
        blurb,
        muPlotRef(
            plotData["tsMonthlyResamples"],
            imageDir
        )
    )
  
    plotData["tsTotalsPerYear"] = {
        "title": "Patients - Annual Additions (2015-{})".format(re.sub(r'\-', '/', end)),
        "plotName": f'{PLOT_PREFIX}',
        "plotMethod": "plotTotalsPerYear",
        "dfcsv": dailiesDFCSV,
        "kargs": {
            "colNameById": vistaNames,
            "start": "2015", 
            "end": end
        }
    }
    blurb = "Finally, a simple __side by side set of totals__ for the last five years shows all systems adding similar numbers, year after year. It also reiterates that the bigger the system, the more patient additions it has - Puget adds up to <span class='yellowIt'>25,000</span> patients per year while Walla Walla adds less than <span class='yellowIt'>4,000</span>."
    mu += "{}\n\n{}\n\n".format(
        blurb,
        muPlotRef(
            plotData["tsTotalsPerYear"],
            imageDir
        )
    )
                                
    return plotData, mu
            
# ############################## DRIVER ################################
            
def main():

    assert sys.version_info >= (3, 6)
    
    if len(sys.argv) > 1:
        if sys.argv[1] == "PLOT":
            makePlots("Common", "Patients")
            return
        raise Exception("Only argument allowed is 'PLOT'")
        
    visn20VistAs = vistasOfVISNByOne("20")

    dirName = "Patient"
    cpd = CommonPatientDirectory(visn20VistAs, CommonUserDirectory(visn20VistAs))

    plotData = {}
    mu = f"""---
layout: default
title: VISN 20 {dirName} Directory
---

# VISN 20 {dirName} Directory

"""

    mu += """The following reports on a VISN 20 Patient Directory built from the patient file (2) of 7 VistAs: Puget Sound [663], Portland [648], Boise [531], Spokane [668], Roseburg [653], White City [692] and Walla Walla [687]. Together they hold <span class='yellowIt'>{:,}</span> distinct patients using <span class='yellowIt'>{:,}</span> patient file records.

Note: Anchorage is the missing VISN 20 system. Its information will be added once the system is made available.

""".format(
        cpd.total(),
        cpd.entryTotal()
    )
    
    ttlPlotData, ttlMU = makeTotalsSection(cpd, visn20VistAs)
    mu += ttlMU
    plotData.update(ttlPlotData)
    
    tsPlotData, tsMU = makeTimeSeriesSection(cpd, visn20VistAs)
    mu += tsMU
    plotData.update(tsPlotData)
        
    plotDir = f'{VISTA_DATA_BASE_DIR}Common/TmpWorking/'
    print(f'Serializing vizData to {plotDir}')
    json.dump(plotData, open(f'{plotDir}plotDataPatients.json', "w"), indent=4)
                
    siteDir = SITE_DIR_TEMPL.format("Common")
    print(f'Serializing Report to {siteDir}')
    open(f'{siteDir}visn20PatientDirectory.md', "w").write(mu) 
        
if __name__ == "__main__":
    main()
    


#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict
from datetime import datetime
import numpy

from fmqlutils.cacher.cacherUtils import SCHEMA_LOCN_TEMPL 
from fmqlutils.reporter.reportUtils import MarkdownTable, reportPercent, reportAbsAndPercent

from ..webReportUtils import SITE_DIR_TEMPL, TOP_MD_TEMPL, keyStats, roundFloat

"""
Basic overview of Data available in system 
"""
def webReportSubReportDataTypes(stationNo):

    def expectedTypeCounts(stationNo):
        try:
            selectTypes = json.load(open(SCHEMA_LOCN_TEMPL.format(stationNo) + "SELECT_TYPES.json"))
        except:
            raise Exception("Can't load SELECT_TYPE.json from {}".format(SCHEMA_LOCN_TEMPL.format(stationNo)))
        expectedCountByType = {}
        ranks = set()
        cnts = []
        for result in selectTypes["results"]:
            if "parent" in result:
                continue
            typ = re.sub(r'\.', '_', result["number"])
            if "count" not in result:
                cnt = 0
            elif re.search(r'\-', result["count"]):
                cnt = -1
            else:
                cnt = int(result["count"]) 
                cnts.append(cnt)
            if cnt not in ranks:
                ranks.add(cnt)
            expectedCountByType[typ] = {"label": result["name"], "count": cnt}
        ranks = sorted(list(ranks), reverse=True)
        percentileThress = []
        for ptile in range(90, 0, -10):
            thres = numpy.percentile(cnts, ptile)
            percentileThress.append((int(thres), ptile))
        for typ in expectedCountByType:
            if expectedCountByType[typ]["count"] == -1:
                continue
            expectedCountByType[typ]["rank"] = ranks.index(expectedCountByType[typ]["count"]) + 1
            if expectedCountByType[typ]["count"] == 0:
                continue
            for percThres in percentileThress:
                if expectedCountByType[typ]["count"] >= percThres[0]:
                    expectedCountByType[typ]["sizePercentile"] = percThres[1]
                    break
        return expectedCountByType
    typeInfoById = expectedTypeCounts(stationNo)

    title = "{} Data Types and Sizes".format(stationNo)
    mu = TOP_MD_TEMPL.format("Subreport Data Types and Sizes", title)
                
    # Reorganize
    labelByTypId = {}
    subReportsByTypId = defaultdict(list)
    for metaInfo in subReportInfo:
        title = metaInfo["title"]
        for typInfo in metaInfo["types"]:
            labelByTypId[typInfo["id"]] = typInfo["label"]
            subReportsByTypId[typInfo["id"]].append((title, typInfo["scope"]))
    
    mu += "## Data Types used for this Report\n\n"
    mu += "<span class='yellowIt'>{:,}</span> types of data are used in the following sections of this report ...\n\n".format(len(labelByTypId))
    
    tbl = MarkdownTable(["Rank", "Type", "Records", "Section"], includeNo=False)
    for typId in sorted(labelByTypId, key=lambda x: typeInfoById[re.sub(r'\.', '_', x)]["rank"]):
        typeInfo = typeInfoById[re.sub(r'\.', '_', typId)]
        tbl.addRow([typeInfo["rank"], "__{}__ ({})".format(labelByTypId[typId], typId), typeInfo["count"], ", ".join(["__{}__ ({})".format(sr[0], sr[1]) for sr in subReportsByTypId[typId]])]) 
    mu += tbl.md() + "\n\n"
    
    mu += "where _ALL_ means all of the records of a type are used while _SO_ means only those records created during the period for which sign on logs exist are examined.\n\n"
    
    mu += "As the report expands more types of data will be examined.\n\n"
    
    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    open(userSiteDir + "subReportDataTypes.md", "w").write(mu)

# ################################# DRIVER #######################
               
def main():

    assert sys.version_info >= (3, 4)

    try:
        stationNo = sys.argv[1]
    except IndexError:
        raise SystemExit("Usage _EXE_ STATIONNO")

    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    if not os.path.isdir(userSiteDir):
        raise SystemExit("Expect User Site to already exist with its basic contents")
        
    TYPES_USED_BY_SUBREPORT = json.load(open(userSiteDir + "subReportInfo.json"))

    webReportSubReportDataTypes(stationNo, TYPES_USED_BY_SUBREPORT)
                 
if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict
from datetime import datetime

from fmqlutils.cacher.cacherUtils import metaOfVistA
from fmqlutils.reporter.reportUtils import MarkdownTable, reportPercent, reportAbsAndPercent

from ..webReportUtils import SITE_DIR_TEMPL, TOP_MD_TEMPL, ensureWebReportLocations

"""
Background:
- Institutions: https://www.va.gov/directory/Guide/fac_list_by_state.cfm?State=OH&dnum=All
... and note that it seems that the default institution IEN from KSP == station number
in all cases.
"""

TEMPL = """## Introduction

Today the Veterans Administration (VA) provides and manages veteran care using <span style='yellowIt'>130</span> deployments of their in-house VistA system. Each VistA is an integrated suite of over one hundred applications sharing a single hierarchal database, FileMan. In deployed VistAs, FileMan holds up to <span style='yellowIt'>2 billion</span> interlinked records of over <span style='yellowIt'>4000</span> different types.

As well as the vast majority of Veteran clinical data, FileMan holds profiles of VA users, financial information, connected device descriptions, interface definitions and usage logs. The integrated and comprehensive nature of FileMan data makes it an ideal target for analysis to inform the migration of veteran care from VistA and into EHRM.

This __\"VistA by the Numbers\"__ report is one of a series, each of which captures the activity of a specific active VA VistA system. Each is generated by a framework for VistA analytics that first caches relevant data from a FileMan and then generates appropriate system descriptions. Such automation ensures common criteria for analysis and the ability to cover all <span style='yellowIt'>130</span> systems deployed in VA even as additional criteria are added to the framework over time.

This report describes __{}__  VistA with station number __{}__. Specifically, it describes a clone cut on __{}__ and housed in the VAEC.

"""

def webReportIntro(stationNo):
    
    meta = metaOfVistA(stationNo)
    
    title = "{} Introduction".format(stationNo)
    mu = TOP_MD_TEMPL.format(title)
    
    mu += TEMPL.format(
        meta.get("name", "VistA"),
        stationNo,
        meta["cutDate"]
    )
    
    mu += "Other defining properties of this clone are ...\n\n"
    
    tbl = MarkdownTable([":Property", ":Value"], includeNo=False)
    for metaPropInfo in [
        ("longName", "Full Name"),
        ("name", "Short Name"),
        ("dateEnteredPostMaster", "Post Master Entered"),
        ("lastSignOnPostMaster", "Post Master Last signon"),
        ("domainName", "Domain Name"),
        ("cio", "CIO"),
        ("iso", "ISO"),
    ]:
        if metaPropInfo[0] not in meta:
            continue
        val = meta[metaPropInfo[0]]
        if isinstance(val, dict):
            val = f'{val["label"]} [{val["id"].split("-")[1]}]'
        tbl.addRow([
            "__{}__".format(metaPropInfo[1]), 
            val
        ])
    if "nameFirstInstall" in meta:
        tbl.addRow([
            "__First Installed Build__",
            "{} ({})".format(meta["nameFirstInstall"], meta["dateLoadedFirstInstall"])
        ])
    mu += tbl.md() + "\n\n"
    
    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    open(userSiteDir + "intro.md", "w").write(mu)
    
# ################################# DRIVER #######################
               
def main():

    assert sys.version_info >= (3, 6)

    try:
        stationNo = sys.argv[1]
    except IndexError:
        raise SystemExit("Usage _EXE_ STATIONNO")

    ensureWebReportLocations(stationNo)

    webReportIntro(stationNo)
    
if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict, Counter
from datetime import datetime

from fmqlutils.reporter.reportUtils import MarkdownTable, reportPercent, reportAbsAndPercent, muBVC
from fmqlutils.cacher.cacherUtils import metaOfVistA
from fmqlutils.typer.reduceTypeUtils import splitTypeDatas, combineSubTypes, checkDataPresent, singleValue, muBVCOfSTProp

from ..webReportUtils import filteredResultIterator, reduce200, reduce4, reduce44, TOP_MD_TEMPL, SITE_DIR_TEMPL, ensureWebReportLocations, keyStats, roundFloat

"""
Overall: Scheduling w/VSE focus

TODO: [1] match webReportEncounter's nuance of noshow(VSE only)/cancelled/active break
for appts and carry it through into tables ie/ can index triple ST by workload and instit 
in detailed report below. [2] prelude to showing % of each type with matching enc - mainly/only active with some active's still being left as orphans. [3] Also can show legacy to VSE matchups where not all legacy has a VSE and vica versa including workload
type.
+
by institution and its stationNo -- seems to take that of first st by institution, st broken by Hospital Location. If locn is wrongly allocated then will end up with wrong station no for Instit in table?

per institution (has station number), showing scheduling based on stop codes. Additional nuance provided by normalized physical location names and, if many locations, forms/templates of locations. Handles case where only legacy (effectively) is working
ie/ VSE little or none for recent 3 month period (ex/ Columbus)

Most obvious missing: link from Appointment (non cancelled) to [Scheduled] Visit
and the normalization of visits (often > 1 resource made) into clean Encounters.

Context: one of the "DSS-based" (https://github.com/vistadataproject/workload/issues/27) reports as Location is key. Part of a flow from Appointment through [Scheduled] Visit (which anchors clinical activity): [1] DSS gathers [2] Appointments and [3] Scheduled Encounters which gather [4] Clinical Activity (record updates for our purposes). Ultimately appropriate, normalized RF data versions should make such reports easy to do.

Short cut: taking DSS from Location, something that changes yearly. Acceptable as only
doing last three months though if that period spans a year, there can be miscues. See
if can trace the "DSS per Location by Year/Period" so accurate.

Two VSE PDFs:
- https://www.va.gov/vdl/documents/Clinical/Scheduling/VSE_GUI_1_5_Technical_Manual.pdf and https://usermanual.wiki/Pdf/VSEGUI15UserGuide.1029341778/view
and
- https://www.va.gov/vdl/documents/Clinical/Scheduling/VSE_GUI_1_4_User_Guide.pdf
"""
def reportAppointmentsVSE(stationNo, vistaName, cutDate):

    allThere, details = checkDataPresent(stationNo, [

        {"fileType": "4", "check": "ALL"},
        {"fileType": "44", "check": "ALL"}, # from ALLRF (temporary)
        
        {"fileType": "409_84", "check": "TYPE"},
        {"fileType": "409_84", "check": "DAY90"},

        {"fileType": "409_831", "check": "ALL"},
        {"fileType": "409_833", "check": "ALL"}

    ])
    if not allThere:
        raise Exception("Some required data is missing - {}".format(details))
            
    _all, sts = splitTypeDatas(stationNo, "409_84") # SDEC Appointments
    
    # expect YR-MTH for starttime but allow for older Year only too
    allSortedStartTime = sorted([(startTime, _all["starttime"]["byValueCount"][startTime]) for startTime in _all["starttime"]["byValueCount"]], key=lambda x: x[0]) 
    sigSortedStartTime = [sv[0] for sv in allSortedStartTime if sv[1] > 500]
    firstSignificantDate = re.sub(r'\-', '/', sigSortedStartTime[0])
    lastSignificantDate = re.sub(r'\-', '/', sigSortedStartTime[-1])
        
    # IMPORTANT: 'expectNotEfficientWalked' important as "starttime" is not ien order
    allDAY90, stsDAY90 = splitTypeDatas(stationNo, "409_84", reductionLabel="DAY90", expectSubTypeProperties=["resource", "#cancel_datetime", "#noshow_datetime"], expectNotEfficientWalked=True) 
    
    # Checking for Columbus and others where VSE may be off (or never on?)
    day90Fraction = float(allDAY90["_total"])/float(_all["_total"])
    if _all["_total"] < 10000 or day90Fraction < .01:
        print("Can't work as either total VSE {:,} too small (< 10K) or the fraction of VSE in last three months has dropped to tiny levels - {}. Exiting VPE Reporting".format(_all["_total"], day90Fraction))
        return
    
    # TODO: move to three way pass down reporting No Show, Cancel, Active
    # ... will fit with Legacy which has two of these three breakdowns.
    stsDAY90Active = [st for st in stsDAY90 if not ("noshow_datetime" in st or "cancel_datetime" in st)]
    # Cancel and not marked noshow
    stsDAY90Cancelled = [st for st in stsDAY90 if "cancel_datetime" in st and "noshow_datetime" not in st] # Cancel only
    # Allows cancel too but noshow takes precedent
    stsDAY90NoShowed = [st for st in stsDAY90 if "noshow_datetime" in st]
    stsDAY90 = combineSubTypes(stsDAY90, ["resource"], forceCountAllProps=True)
    
    institInfosByIEN = reduce4(stationNo)
    hlInfosByIEN = reduce44(stationNo, useRF=False)
    resourceInfosByIEN = reduce409_831(stationNo)
                
    # ENFORCE/SEPARATE: Making a statement that only doing Location resources - enforce it
    def enforceResourcesAllLocation(stsDAY90):
        nstsDAY90 = []
        for st in stsDAY90:
            try:
                resourceRef = singleValue(st, "resource")
            except:
                # in 663, say NOVALUE meant 'note' set to CX IN ERROR
                print("** Warning: expected st's of 409_84 to have resource set but {} with {:,} appointments didn't.".format(st["_subTypeId"], st["_total"]))
                continue
            resourceIEN = re.search(r'\-(\d+)\]', resourceRef).group(1)
            resourceInfo = resourceInfosByIEN[resourceIEN]
            if resourceInfo["source_resource_class"] != "LOCATION":
                raise Exception("Expect all resources with appointments to be LOCATION")
            nstsDAY90.append(st)
        return nstsDAY90
    stsDAY90 = enforceResourcesAllLocation(stsDAY90)
        
    # Make explicit the hospital location from the Location Resource 
    for st in stsDAY90:
        resourceRef = singleValue(st, "resource")
        resourceIEN = re.search(r'\-(\d+)\]', resourceRef).group(1)
        resourceInfo = resourceInfosByIEN[resourceIEN]
        st["_hospital_location"] = resourceInfo["hospital_location"]   
    stsDAY90 = enhanceHLocSTsByLOC(institInfosByIEN, hlInfosByIEN, stsDAY90)
                
    # TMP: for now, keeping separate from norm of 44 so can work incrementally     
    try:
        PLOCMAP = json.load(open("MetaPerStation/{}/physicalLocationsNorm.json".format(stationNo)))["map"]
    except:
        PLOCMAP = None        
                
    # ############################ MU REPORT ################################
            
    mu = """VistA Scheduling Enhancements (VSE) is the package used for appointment creation and tracking. Despite the word \"Enhancements\", VSE is a new package that superceeds an older scheduling package.
    
"""
    
    muProviderTotal = "Only <span class='yellowIt'>{}</span> of these _Location Appointments_ specify a provider".format(reportAbsAndPercent(allDAY90["provider"]["count"], allDAY90["_total"])) if "provider" in allDAY90 else "No appointments specify a provider"
    
    # REM: three month scheduled == occur in this period ie/ not made in it but meant to
    # take place.
    mu += """{} records <span class='yellowIt'>{:,}</span> VSE appointments, the first in significant numbers in {}, the latest in {}. 
    
For the 90 days before this VistA copy was cut (_{}_), the System shows <span class='yellowIt'>{}</span> appointments were scheduled for <span class='yellowIt'>{:,}</span> patients and <span class='yellowIt'>{:,}</span> \"Resources\". <span class='yellowIt'>{}</span> of these appointments were active, <span class='yellowIt'>{}</span> were formally cancelled and <span class='yellowIt'>{}</span> were \"no shows\".
    
The software allows a \"Resource\" to be a provider or a location but in this system, all resources with appointments from 2019 are locations - in other words, appointments are only made for locations (\"Location Appointment\"). {}.

""".format(
        "__{} [{}]__".format(vistaName, stationNo),
        _all["_total"],
        firstSignificantDate,
        lastSignificantDate,
        
        cutDate,
        reportAbsAndPercent(allDAY90["_total"], _all["_total"]),
        allDAY90["patient"]["rangeCount"],
        allDAY90["resource"]["rangeCount"],
        reportAbsAndPercent(
            sum(st["_total"] for st in stsDAY90Active),
            allDAY90["_total"]
        ),
        reportAbsAndPercent(
            sum(st["_total"] for st in stsDAY90Cancelled),
            allDAY90["_total"]
        ),
        reportAbsAndPercent(
            sum(st["_total"] for st in stsDAY90NoShowed),
            allDAY90["_total"]
        ),
        
        muProviderTotal # very unlikely: say in Columbus as moved on from VSE
    )
            
    mu += muByInstitutions(allDAY90, stsDAY90, plocMap=PLOCMAP)
    
    mu += muAboutWorkLocation() 
    
    """
    TODO: markup or check "length of appt" and see date appt made vs
    split on appt time ie/ where one is == or < or ... (same for 44_003 below)
    
    Also appointment_type (which was removed from generic)
    """
    propsPresenceToCount=["checkin", "checkout", "noshow", "consult_link", "provider", "cancel_datetime"]
    mu += muByStopCodes(stationNo, stsDAY90, plocMap=PLOCMAP, propsPresenceToCount=propsPresenceToCount)
                    
    # #################### FINAL WRITE OUT ###########################
    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    print("Serializing VSE Appointment Report to {}".format(userSiteDir))
    if not os.path.isdir(userSiteDir):
        raise Exception("Expect User Site to already exist with its basic contents")
    title = "VSE Appointments {}".format(stationNo)
    mu = TOP_MD_TEMPL.format(title, "VSE Appointments") + mu
    open(userSiteDir + "appointmentsVSE.md", "w").write(mu)
    
# ############################## Legacy Form Appointments 44_003 #######################

"""
In form, this follows VPE (409_84) reports except where Legacy 44.003 has its own
properties (consult ref)

TODO: add from type of 44_003 here ala for 409_84 above
"""
def reportAppointmentsLegacy44(stationNo, vistaName, cutDate):

    allThere, details = checkDataPresent(stationNo, [

        {"fileType": "4", "check": "ALL"},
        {"fileType": "44", "check": "ALLRF"},
        {"fileType": "44_003", "check": "TYPE"},
        {"fileType": "44_003", "check": "DAY90"}

    ])
    if not allThere:
        raise Exception("Some required data is missing - {}".format(details))

    institInfosByIEN = reduce4(stationNo)
    hlInfosByIEN = reduce44(stationNo)

    LOCN_PROP = "appointment_container_"

    _all, sts = splitTypeDatas(stationNo, "44_003")
        
    # expect YR-MTH for starttime but allow for older Year only too
    allSortedADT = sorted([(adt, _all["appointment_date_time"]["byValueCount"][adt]) for adt in _all["appointment_date_time"]["byValueCount"]], key=lambda x: x[0]) 
    sigSortedADT = [sv[0] for sv in allSortedADT if sv[1] > 500]
    firstSignificantDate = re.sub(r'\-', '/', sigSortedADT[0])
    lastSignificantDate = re.sub(r'\-', '/', sigSortedADT[-1])

    allDAY90, stsDAY90 = splitTypeDatas(stationNo, "44_003", reductionLabel="DAY90", expectSubTypeProperties=["appointment_container_", "#appointment_cancelled"], expectNotEfficientWalked=True)     
    stsDAY90Active = [st for st in stsDAY90 if re.search(r'ABSENT', st["_subTypeId"])]
    # TODO: eventually will only do non cancelled as cuts out ZZ's too
    stsDAY90Cancelled = [st for st in stsDAY90 if re.search(r'PRESENT', st["_subTypeId"])]
    # TMP: for now, only want resource based sts
    stsDAY90 = combineSubTypes(stsDAY90, ["appointment_container_"], forceCountAllProps=True)
    
    # TMP: for now, keeping separate from norm of 44 so can work incrementally     
    try:
        PLOCMAP = json.load(open("MetaPerStation/{}/physicalLocationsNorm.json".format(stationNo)))["map"]
    except:
        PLOCMAP = None  
        
    for st in stsDAY90:
        if LOCN_PROP not in st:
            continue
        hlRef = singleValue(st, LOCN_PROP)
        st["_hospital_location"] = hlRef
        # Note: not enforce all locs have stop_code (warned below)
        hlInfo = hlInfosByIEN[re.search(r'(\d+)\]', hlRef).group(1)]
    stsDAY90 = enhanceHLocSTsByLOC(institInfosByIEN, hlInfosByIEN, stsDAY90)
    
    # ############################ MU REPORT ################################
            
    mu = "VistA Scheduling (Legacy) was the only mechanism used for managing appointments in VistA before VistA Scheduling Enhancements (VSE).\n\n"   
    
    mu += """{} records <span class='yellowIt'>{:,}</span> legacy appointments, the first in significant numbers in {}, the latest in {}. 
    
For the 90 days before this VistA copy was cut (_{}_), the System shows <span class='yellowIt'>{}</span> appointments were scheduled for <span class='yellowIt'>{:,}</span> patients and <span class='yellowIt'>{:,}</span> \"Locations\". <span class='yellowIt'>{}</span> of these appointments were active, <span class='yellowIt'>{}</span> were formally cancelled.

""".format(
        "__{} [{}]__".format(vistaName, stationNo),
        _all["_total"],
        firstSignificantDate,
        lastSignificantDate,
        
        cutDate,
        reportAbsAndPercent(allDAY90["_total"], _all["_total"]),
        allDAY90["patient"]["rangeCount"],
        allDAY90[LOCN_PROP]["rangeCount"],
        reportAbsAndPercent(
            sum(st["_total"] for st in stsDAY90Active),
            allDAY90["_total"]
        ),
        reportAbsAndPercent(
            sum(st["_total"] for st in stsDAY90Cancelled),
            allDAY90["_total"]
        )
    )
            
    mu += muByInstitutions(allDAY90, stsDAY90, plocMap=PLOCMAP)
    
    mu += muAboutWorkLocation() 
    
    """
    TODO: markup or check "length of app't" and see date_appointment_made vs
    split on appointment_date_time ie/ where one is == or < or ...
    """
    propsPresenceToCount=["checked-in", "checked_out", "appointment_cancelled", "consult_link"]
    mu += muByStopCodes(stationNo, stsDAY90, plocMap=PLOCMAP, propsPresenceToCount=propsPresenceToCount)
            
    # #################### FINAL WRITE OUT ###########################
    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    print("Serializing Legacy Appointment Report to {}".format(userSiteDir))
    if not os.path.isdir(userSiteDir):
        raise Exception("Expect User Site to already exist with its basic contents")
    title = "Legacy Appointments {}".format(stationNo)
    mu = TOP_MD_TEMPL.format(title, "Legacy Appointments") + mu
    open(userSiteDir + "appointmentsLegacy.md", "w").write(mu)

# ################################# Common MU's/ HL Enhance ###################

"""
Focus is [1] per institution and [2] stop codes as [3] work locations are combination concepts. Need to report on both [A] VPE and [B] Legacy 44.003 (and maybe per patient too)
"""
def muAboutWorkLocation(): # background blurb shared
    
    mu = "In VistA, location (\"Work Location\") represents the combination of [i] a type of work in [ii] a specific institution performed by [iii] a particular provider, group of providers or class of provider. Appointments and Visits are organized around these Work Locations.\n\n"
    
    mu += "The work type of each Work Location is formally specified with a VA-custom \"stop code number\", the key designator for VA's _Decision Support System_ (DSS) ...\n\n"
    
    # "DSS calculates the actual cost to provide patientcare services"
    mu += "> VHA collects workload data that supports the continuity of patient care, resource allocation, performance measurement, quality management, and third-party collections. Decision Support System (DSS) Identifiers assist VA medical centers in defining workload, which is critical for costing purposes. DSS Identifiers are used to identify workload for all outpatient encounters, inpatient appointments in outpatient clinics, and inpatient billable professional services. They also serve as guides to select DSS outpatient department structures.\n\n"
    
    mu += "This approach downplays physical location as one physical location may host many Work Locations. Even when specified in the system, physical location is a freeform text string that varies greatly in form and quality - it has been somewhat normalized in the tables below.\n\n"
    
    mu += "The use of Work Locations for appointments makes _stop code_ the best mechanism for describing and typing appointments.\n\n"
    
    return mu
    
"""
By Institution summary 

Expects: _institution, _stop_code_number added to st's
"""
def muByInstitutions(_all, sts, plocMap=None): # TODO: map ploc ala sc mu below

    _allTotal = sum(st["_total"] for st in sts) # have _all too for patient range total but TODO, will move off that.
    stsByInstitution = defaultdict(list)
    stsNoInstitution = [] # again
    stByPhyLocByInstit = defaultdict(lambda: defaultdict(list))
    patientsByInstitution = defaultdict(set)
    apptCountByInstitution = Counter()
    apptTypeCountByInstitution = defaultdict(lambda: Counter()) # only 409_84
    for st in sts:
        if "_institution" not in st:
            stsNoInstitution.append(st)
            continue
        instit = st["_institution"]
        stsByInstitution[st["_institution"]].append(st)
        apptCountByInstitution[instit] += st["_total"]
        if "byValueCount" not in st["patient"]:
            raise Exception("Reduction of st didn't enforce count 'patient' so no byValueCount")
        patientsByInstitution[st["_institution"]] |= set(st["patient"]["byValueCount"])
        if "_physical_location" in st:
            pl = st["_physical_location"].upper()
            stByPhyLocByInstit[instit][pl].append(st)
        if "appointment_type" in st: # 409_84 only
            bvc = st["appointment_type"]["byValueCount"]
            for v in bvc:
                apptTypeCountByInstitution[instit][v] += bvc[v] 
                
    def weekdayWeekendCounts(st, wdweCounts):
        if not ("_createDateProp" in st and st["_createDateProp"] in st):
            return wdweCounts
        createDatePropInfo = st[st["_createDateProp"]]
        if "byWeekDay" not in createDatePropInfo:
            return wdweCounts
        wdweCounts["WD"] += sum(createDatePropInfo["byWeekDay"][wd] for wd in createDatePropInfo["byWeekDay"] if int(wd) < 5)
        wdweCounts["WE"] += sum(createDatePropInfo["byWeekDay"][wd] for wd in createDatePropInfo["byWeekDay"] if int(wd) > 4)
        return wdweCounts

    cols = [":Institution", "Locns", ":Locn Prefixes", "Phy Locns", "Stop Codes", "PTs", "Appts", "Weekend"]
    if len(apptTypeCountByInstitution):
        cols.append(":Appointment Types")
    tbl = MarkdownTable(cols)
    for instit in sorted(apptCountByInstitution, key=lambda x: apptCountByInstitution[x], reverse=True):
        name = re.sub(r' +$', '', instit.split(" [")[0])
        nameMU = "__{}__ [{}]".format(name, st["_station_number"]) if "_station_number" in st else "__{}__".format(name)
        byStopCodeCount = Counter()
        hlocPrefixes = Counter()
        wdweCounts = Counter()
        for st in stsByInstitution[instit]:
            if "_stop_code_number" not in st:
                continue
            byStopCodeCount[st["_stop_code_number"]] += 1
            hlocPrefixes[st["_prefix"] if "_prefix" in st else "NO PREFIX"] += 1
            wdweCounts = weekdayWeekendCounts(st, wdweCounts)
        weekendMU = reportPercent(wdweCounts["WE"], apptCountByInstitution[instit]) if wdweCounts["WE"] != 0 else ""
        row = [
            nameMU,
            len(stsByInstitution[instit]),
            muBVC(hlocPrefixes, totalOver=7), # may cut down (more a QA)
            len(stByPhyLocByInstit[instit]) if instit in stByPhyLocByInstit else "",
            len(byStopCodeCount),
            reportAbsAndPercent(len(patientsByInstitution[instit]), _all["patient"]["rangeCount"]),
            reportAbsAndPercent(apptCountByInstitution[instit], _allTotal),
            weekendMU
        ]
        if len(apptTypeCountByInstitution):
            row.append(muBVC(apptTypeCountByInstitution[instit]))
        tbl.addRow(row)
    mu = "<span class=\"yellowIt\">{:,}</span> \"Institutions\" supported by this system have Appointments spread over <span class=\"yellowIt\">{:,}</span> locations with <span class=\"yellowIt\">{:,}</span> stop codes ...\n\n".format(
        len(apptCountByInstitution),
        sum(len(stsByInstitution[instit]) for instit in stsByInstitution),
        len(set(st["_stop_code_number"] for st in sts if "_stop_code_number" in st))
    )
    mu += tbl.md() + "\n\n"
    
    # IMPORTANT: rare rare not to have stop code. Usually won't have _institution.
    if len(stsNoInstitution):
        mu += "__Note__: <span class=\"yellowIt\">{:,}</span> location(s) with <span class=\"yellowIt\">{:,}</span> appointments either don't specify an institution or one can't be derived from its name. <span class=\"yellowIt\">{}</span> of these are deleted locations, ones whose names begin with _ZZ_.\n\n".format(
            len(stsNoInstitution),
            sum(st["_total"] for st in stsNoInstitution),
            reportAbsAndPercent(
                sum(1 for st in stsNoInstitution if "_hospital_location" in st and re.match(r'ZZ', st["_hospital_location"])),
                len(stsNoInstitution)
            ) 
        )
    
    return mu
    
"""
Stop Code as Anchor - [1] separate out the non 'in person' appts from [2] TELEPHONE 
etc for the primary institutions and then all others.
    
See: https://www.va.gov/vdl/documents/Financial_Admin/Decision_Supp_Sys_(DSS)/dss_ug.pdf, http://vistadataproject.info/demo2/ and http://vistadataproject.info/demo2/location (need to rework these)

NOTE: no institution (and rarer no stop_code) sts are passed in and used to count the
total. There are counted in the Note about ("appointments either don't specify an institution or one can't be derived from its name"). It is not worth counting them again.
    
TODO Structure:
- ala webReportEncounter, distinguish cancelled, no show, active (=> will need noshow in 409_84 breakdown; Legacy doesn't do noshow)
    
TODO Overall:
- more on those others (TELEPHONE etc)
- TELEHEALTH -- go beyond stops
- PRIMARY CARE broken out as secondaries won't do anything ... need more
- GROUPs of Stop Codes    
"""
def muByStopCodes(stationNo, sts, plocMap=None, propsPresenceToCount=[]):

    # FUTURE: consider grouping stop codes together
    SC_GROUPS = [
        ["ORTHO JOINT SURG", "PODIATRY"], # Building 1e SPECIALITY for this
        ["OPHTHALMOLOGY", "OPTOMETRY"], # Building 40 ... expect SPO EYE for first twos
        ["BROS (BLIND REHAB O P SPEC)", "PHYSICAL THERAPY", "CHIROPRACTIC CARE", "OCCUPATIONAL THERAPY", "WHEELCHAIR & ADVAN MOBILITY", "RECREATION THERAPY SERVICE", "PROSTHETICS ORTHOTICS"], # Building 31
        ["MH INTGRTD CARE GRP", "MH INTGRTD CARE IND"], # group these but not others?
        ["PSYCHOLOGICAL TESTING", "MH INTGRTD CARE IND", "PTSD - INDIVIDUAL"], # B40, BHS (not SUBSTANCE USE DISORDER IND)
        ["SUBSTANCE USE DISORDER IND", "SUBSTANCE USE DISORDR GRP"],
        
        ["MENTAL HEALTH CLINIC - IND", "MENTAL HEALTH CLINIC-GROUP", "MH INTGRTD CARE IND", "MH TEAM CASE MANAGEMENT", "PSYCHOLOGICAL TESTING", "PTSD - GROUP", "PTSD - INDIVIDUAL", "SUBSTANCE USE DISORDR GRP"]
    ]
    
    # FUTURE: more expansion for Workload Location names
    MED_ACRONYMS = {
    
        "DI": "DIAGNOSTIC IMAGING",
        "BHS": "BEHAVIORAL HEALTH SERVICES"
        
    }
                                
    def muAppointmentsOfStopCodes(instit, stsOfSC, plocMap=None, propsPresenceToCount=[], showLocations=False):
        def dashNBVC(sts, prop, valMap=None):
            cntr = Counter()
            for st in stsOfSC[sc]:
                if prop not in st:
                    val = "*NOT SET*"
                elif re.match(r'\_', prop):
                    val = re.sub(r' +$', '', st[prop].upper())
                    if valMap and val in valMap:
                        val = valMap[val]
                    cntr[val] += st["_total"]
                else:
                    for val in st[prop]["byValueCount"]:
                        cntr[val.split(" [")[0]] += st[prop]["byValueCount"][val]
            return cntr 
        def bvcSTSs(sts, prop):
            cntr = Counter()
            for st in sts:
                for val in st[prop]["byValueCount"]:
                    cntr[val.split(" [")[0]] += st[prop]["byValueCount"][val]
            return cntr
        def bvcPropsPresence(sts, props):
            cntr = Counter()
            for st in sts:
                for prop in props:
                    if prop not in st:
                        continue
                    cntr[prop] += st[prop]["count"]
            return cntr
        def countLocnPrefixes(sts):
            cntrAppear = Counter()
            entries = []
            for st in sts:
                locn = re.sub(r'^ +', '', re.sub(r' +$', '', st["_hospital_location"].split(" [")[0]))
                pieces = locn.split(" ")
                entry = [[], st["_total"]]
                for i in range(1, len(pieces) + 1):
                    k = " ".join(pieces[0:i])
                    cntrAppear[k] += 1
                    entry[0].append(k)
                entries.append(entry)
            cntr = Counter()
            for entry in entries: # reduce entries and count 
                ks = entry[0]
                if len(ks) >= 3 and cntrAppear[ks[2]] > 1 and not re.search(r' (\d+|[A-Z])$', ks[2]):
                    cntr[ks[2]] += entry[1]
                    continue
                if len(ks) >= 2 and cntrAppear[ks[1]] > 1:
                    cntr[ks[1]] += entry[1]
                    continue
                cntr[ks[-1]] += entry[1] 
            return cntr
        tbl = MarkdownTable([":Stop Code", "Locns", ":Location [Roots]", ":Physical Location(s)", "PTs", "Appts", ":Props"])
        scs = sorted(stsOfSC, key=lambda x: sum(st["_total"] for st in stsOfSC[x]), reverse=True)
        rplocMap = {}
        if plocMap:
            for ploc in plocMap:
                for rploc in plocMap[ploc]:
                    rplocMap[rploc] = ploc
        for sc in scs:
            locns = set(st["_hospital_location"] for st in stsOfSC[sc]) 
            phyLocCntr = dashNBVC(stsOfSC[sc], "_physical_location", rplocMap)     
            tbl.addRow([
                "__{}__".format(re.sub(r'\_', ' ', sc.split(" [")[0])),
                muBVC(dashNBVC(stsOfSC[sc], "_hospital_location")) if showLocations else len(locns),
                muBVC(countLocnPrefixes(stsOfSC[sc])),
                muBVC(phyLocCntr) if len(phyLocCntr) else "",
                len(bvcSTSs(stsOfSC[sc], "patient")),
                reportAbsAndPercent( # % of these appts (not total overall instits)
                    sum(st["_total"] for st in stsOfSC[sc]),
                    sum(st["_total"] for sc in scs for st in stsOfSC[sc])
                ),
                re.sub(r'\_', ' ', muBVC(bvcPropsPresence(stsOfSC[sc], propsPresenceToCount)))
            ])
        return tbl.md() + "\n\n"
        
    _allTotal = sum(st["_total"] for st in sts)
    
    stsWSC = [st for st in sts if "_stop_code_number" in st and "_institution" in st]
    
    # Note: for now, not explicitly noting these are noted above in Note with
    # 'appointments either don't specify an institution or one can't be derived ...'
    # ... may reconsider whether _allTotal should have them or not but not worth excluding
    # now as [1] small in # and [2] only showing in-person appointments in tables so
    # they are a subset anyhow
    notCountedTotal = sum(st["_total"] for st in sts if not ("_stop_code_number" in st and "_institution" in st))
    def catagST(st):
        for mtch, catag in [
            (r'(TELEPHONE|PHONE)', "TELEPHONE"),
            (r'(COMMUNITY CARE CONSULT)', "CC"),
            (r'(ADMIN PAT ACTIVTIES)', "ADMIN"),
            (r'(HOME)', "HOME")
        ]:
            if re.match(mtch, st["_stop_code_number"]):
                return catag
        return ""
    stsByCategory = defaultdict(list)
    for st in stsWSC:
        catag = catagST(st)
        if catag:
            stsByCategory[catag].append(st)
            continue
        """
        TODO: doesn't take them ALL -- if TELEHEALTH in physical_location takes
        others w/o a secondary code
        """
        if "_credit_stop_code" in st and re.search(r' TH ', st["_credit_stop_code"]):
            stsByCategory["TELEHEALTH"].append(st)
            continue
        stsByCategory["INPERSON"].append(st)

    mu = "## By Stop Code\n\n"
    # May add: The range of states transitioned by appointments from checkin to checkout seems to vary with work type.
    mu += """The system holds <span class='yellowIt'>{}</span> in-person, clinical appointments. There are also <span class='yellowIt'>{}</span> _TELEPHONE_, <span class='yellowIt'>{}</span> _COMMUNITY CARE_, <span class='yellowIt'>{}</span> _in home care_, <span class='yellowIt'>{}</span> _Telehealth_ and <span class='yellowIt'>{}</span> _administrative appointments_.
    
The tables below break down appointments by institution. _Location_ counts the root or base of location names as the number of individual locations in larger institutions get unmanageable for a table and _Physical Location_ shows the free text names given for buildings and rooms. _Props_ count how many appointments have defining properties such as _"cancel datetime"_ or _checked out_. 
    
""".format(
        reportAbsAndPercent(
            sum(st["_total"] for st in stsByCategory["INPERSON"]),
            _allTotal
        ),
        reportAbsAndPercent(
            sum(st["_total"] for st in stsByCategory["TELEPHONE"]),
            _allTotal
        ),
        reportAbsAndPercent(   
            sum(st["_total"] for st in stsByCategory["CC"]),
            _allTotal
        ),
        reportAbsAndPercent(   
            sum(st["_total"] for st in stsByCategory["HOME"]),
            _allTotal
        ),
        # Undercounted as secondary not enough to break out and NO PRIMARY code
        # ... could use physical location matches but didn't for now
        reportAbsAndPercent(   
            sum(st["_total"] for st in stsByCategory["TELEHEALTH"]),
            _allTotal
        ),
        reportAbsAndPercent(   
            sum(st["_total"] for st in stsByCategory["ADMIN"]),
            _allTotal
        )
    )
    inPersonSTsByInstitBySC = defaultdict(lambda: defaultdict(list))
    stationNumberOfInstit = {}
    for st in stsByCategory["INPERSON"]:
        instit = st["_institution"]
        inPersonSTsByInstitBySC[instit][st["_stop_code_number"]].append(st)
        stationNumberOfInstit[instit] = st["_station_number"] if "_station_number" in st else ""
    for instit in sorted(inPersonSTsByInstitBySC, key=lambda x: sum(st["_total"] for sc in inPersonSTsByInstitBySC[x] for st in inPersonSTsByInstitBySC[x][sc]), reverse=True):
        mu += """Institution _{} [{}]_ uses <span class='yellowIt'>{:,}</span> stop codes for <span class='yellowIt'>{}</span> in-person appointments. Where there are many Work Locations, only the root (first pieces) of those locations are shown,  ...
        
""".format(
            re.sub(r' +$', '', instit.split(" [")[0]), 
            stationNumberOfInstit[instit],
            len(inPersonSTsByInstitBySC[instit]),
            reportAbsAndPercent(sum(st["_total"] for sc in inPersonSTsByInstitBySC[instit] for st in inPersonSTsByInstitBySC[instit][sc]), _allTotal)
        )
        mu += muAppointmentsOfStopCodes(instit, inPersonSTsByInstitBySC[instit], plocMap, propsPresenceToCount)
        
    # TODO: more on Telehealth etc + break that out more as still inside as codes aren't enough
    
    return mu
    
"""
Pattern: enhance sts with _ props from HLOC and Instit

Expect: _hospital_location to be in st
Copies key HLOC props into st along with station_number of Institution for ease of 
mu - makes sure same props for both 409_84 and 44_003 reductions (and more).
"""
def enhanceHLocSTsByLOC(institInfosByIEN, hlInfosByIEN, sts, locationProp="_hospital_location"):
    def addHLPropsToApptST(st, hlInfo):
        pieces = re.split(r'[ \-\/]', hlInfo["label"])
        if len(pieces) > 1:
            st["_prefix"] = pieces[0]
        for prop in ["stop_code_number", "credit_stop_code", "physical_location", "patient_friendly_name", "abbreviation", "provider", "institution"]:
            if prop in hlInfo:
                st["_" + prop] = hlInfo[prop]
        return st
    for st in sts:
        if locationProp not in st:
            continue
        hlIEN = re.search(r'(\d+)\]', st[locationProp]).group(1)
        try:
            hlInfo = hlInfosByIEN[hlIEN]
        except:
            print("Couldn't find {} for {} - probably hl red is incomplete".format(hlIEN, locationProp))
            raise
        addHLPropsToApptST(st, hlInfo)
        if "institution" not in hlInfo:
            continue
        # 541 has . IENs for institution pointer
        institIEN = re.search(r'([\d\.]+)\]$', hlInfo["institution"]).group(1)
        try:
            institInfo = institInfosByIEN[institIEN]
        except:
            raise
        if "station_number" in institInfo:
            st["_station_number"] = institInfo["station_number"]
    return sts

# ################################# Reduce/Lookup #################

"""
Break Resource 409_831 - VPE has indirection for NEW PERSON (200) and
HOSPITAL LOCATION (44) appointments.
"""
def reduce409_831(stationNo):
    from fmqlutils.cacher.cacherUtils import TMPWORKING_LOCN_TEMPL
    jsnFl = TMPWORKING_LOCN_TEMPL.format(stationNo) + "vpeResourceReduction.json"
    try:
        infosByIEN = json.load(open(jsnFl))
    except:
        pass
    else:
        return infosByIEN
    infosByIEN = {}
    for resource in filteredResultIterator(stationNo, "409_831"):
        missingMands = set(["entered_by_user", "date_time_entered", "resource_type"]) - set(resource)
        if len(missingMands):
            raise Exception("Resource 409_831 missing mandatory properties - {}".format(missingMands))
        red = {"_ien": resource["_id"].split("-")[1]}
        infosByIEN[resource["_id"].split("-")[1]] = red
        for prop in resource:
            if re.match(r'_', prop) or prop in ["label", "type"]:
                continue
            if prop == "resource_type": # VPTR to lead to two
                red["source_resource"] = "{} [{}]".format(resource[prop]["label"], resource[prop]["id"]) # no split as VPTR
                trt = resource[prop]["id"].split("-")[0] 
                # Note: perhaps ??? TODO 200's do too?
                if trt == "44" and "hospital_location" not in resource:
                    raise Exception("Expected resource 44's to have hospital location")
                if trt == "44":
                    red["source_resource_class"] = "LOCATION" 
                elif trt == "200":
                    red["source_resource_class"] = "USER"
                else: 
                    red["source_resource_class"] = "SDEC ADDITIONAL" # none in SPOKANE
                continue
            if isinstance(resource[prop], dict):
                if "value" in resource[prop]:
                    red[prop] = resource[prop]["value"]
                    continue
                if "id" in resource[prop]:
                    red[prop] = "{} [{}]".format(resource[prop]["label"], resource[prop]["id"].split("-")[1]) 
                    continue
                raise Exception("Dict Valued Prop not date or ref")
            red[prop] = resource[prop]        
    print("Dumping 409_831 reduction for first time - {:,} LOCATION's, {:,} USER's, {:,} SDEC ADDITIONALs".format(
        sum(1 for ien in infosByIEN if infosByIEN[ien]["source_resource_class"] == "LOCATION"),
        sum(1 for ien in infosByIEN if infosByIEN[ien]["source_resource_class"] == "USER"),            
        sum(1 for ien in infosByIEN if infosByIEN[ien]["source_resource_class"] == "SDEC ADDITIONAL")            
    ))
    json.dump(infosByIEN, open(jsnFl, "w"), indent=4)
    return infosByIEN
        
# ################################# DRIVER #######################

def main():

    assert sys.version_info >= (3, 6)

    try:
        stationNo = sys.argv[1]
    except IndexError:
        raise SystemExit("Usage _EXE_ STATIONNO")

    ensureWebReportLocations(stationNo)
    
    meta = metaOfVistA(stationNo)
    vistaName = "VistA" if "name" not in meta else meta["name"]
    cutDate = meta["cutDate"]

    reportAppointmentsVSE(stationNo, vistaName, cutDate)
    reportAppointmentsLegacy44(stationNo, vistaName, cutDate)

if __name__ == "__main__":
    main()




#!/usr/bin/env python
# -*- coding: utf8 -*-

#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import io
import json
from collections import defaultdict, Counter
from datetime import datetime

import pandas as pd

from fmqlutils import VISTA_DATA_BASE_DIR
from fmqlutils.cacher.cacherUtils import SCHEMA_LOCN_TEMPL 
from fmqlutils.reporter.reportUtils import MarkdownTable, reportPercent, reportAbsAndPercent
from fmqlreports.webReportUtils import SITE_DIR_TEMPL, TOP_MD_TEMPL, keyStats, roundFloat, vistasOfVISNByOne

from fmqlreports.webReportUtils import muPlotRef, makePlots 
from userDirectory import CommonUserDirectory
        
"""
QUICK TODO: 
- reorder the render of user totals in time series (last graph) so biggest appears first ie/ key has PUG, POR ... colNameById needs to be ordered

Goal for directories in general - user ala patient with one, move id checks there but
first need to see what's there now. Federated Directory => mismatch possible, dup,
possible

REM: want ALL User (and Patient?) fixed up and cleaned up (individ and dir) by Labor Day
[one of three key areas inc IFC/TeleR + Data Shape/Sizes]

TODO: and in parallel, improve and make visuals generic
- [ ] make clear the time span of the individual vista (ie/ cut date!) --- want one cut date for directory ie/ cut off all to time of oldest VistA
- [ ] remote vs local user! KEY! --- feeds WWW etc mostly elsewhere (all 4!)
  ... correlate with patients in all? get into 'scatters?'
  (want the "user gone by Labor day")
- [ ] secid growing user (new dups etc) ... "SECID-use" issue! ... 
- [ ] term/demise (user and patient (more))
- [ ] FIX to end date of earliest for data ... as don't want SPO (with extra month) over POR etc in 2020. Leave start date as is.
- [ ] look on graphs -- | black on X and Y axes so clearer
- [ ] Link VISN 20 (VISN 20 https://www.va.gov/directory/guide/region.asp?ID=1020) Directories -- shared care across hospitals and clinics ie/ can treat as unit! ... ALSO may find some are more stand-alone (anchrorage doesn't matter? ... gotta see visits froms)
- New Plot
  - [ ] Scatter Plot use ---> Demise vs Patients added per month ie/ two tracked
  - [ ] Patient age at addition -- does this vary? [age of live patients at cut time]
  (see scatter: https://towardsdatascience.com/5-amazing-tips-for-data-visualization-990fdb19c396)
- [I] time series left
  - "ALL" total ie/ first date of any user
    ... change to serialize CSV at same time in directory using 
  <----- doing user/ seems to make sense but make sure ... main place is totals (NEED TO LOOK INTO IF TRULY NEW ETC ... ie/ DO THE TELEREADER ONLY DIRECTORY USER REPORT and then get back to this)
- [ ] go back to two first months of year for dir as three doesn't stand out right ???
- [ ] BIG And KEY -- HOME PATIENT, HOME USER <---- tell how ... do a pass alg ie/ user => created explicitly
- [user only] no creator vs creator [Auto Entry Addition] ... telling on 2006 or 2019? --- interplay with TERMINATION (a lot a lot in Portland ie/ show contrast)
- [user + patient] 'term' and 'demise' ...
- missing date fill in ie/ evens (got in Patient, add to User and must graph records w/o - extrapolate by surroundings? ie/ guessed date) ---- really only a Patient Problem (seems mainly ok for user)
- automate start-end off VistA image start end (and better doc in md) ie/ automate the ranges (put into dir entries?) ... fix rogues?
  REM: por, www oldest at 3/25 => why do feb back ... cut time series back to there too
  [means for report can handle rolling start-end]
- Variation: Registration Time Delay for studies ie/ 1 patients registered due to studies and two, delay! [studyId, studyDay, patient create day acq, patient create day read] ie/ gets to Patient being in > 1 system and how they get there!

[Context: want fuller Sankey for IFC, TeleR completed in this time frame too]

TODO  NEXT: bigger
- BETTER SPLIT of PATIENT/ USER vs records with fan out of patient, user (first created somewhere) to various systems. ie/ pitch is normalized with (in case of user) 'wasteful'
VistA copies. IE/ FEDERATED DIRECTORIES have issues ... - records flow to ... first
user or patient.
- patient and why ... for IFC? for appt? ie/ what's the trigger?
- feed directory reports back into individual system reports
"""

# ###################################### Totals  ############################

# ... put in line about USER, PATIENT total as a stand out ala NYT graph ie/ all
# lines but that line are greyed
# ... blurb on Directory => Users may have > 1 record in a VistA (misplaced duplication will be a section) ... lifecycle management too (with last logins from anywhere?)
# GOAL: centralize user mgmt to ease migration and management [ 1 of a # for dec]

PLOT_PREFIX = "visn20User"

def makeTotalsSection(cd, vistaNames, imageDir="Images"):
                       
    plotData = {}
                       
    mu = "## User Breakdown\n\n"
               
    countsFromTotal = {
        "__entityName": "Users",
    
        "__total": cd.total(),
        
        "> 1 VistA": cd.gt1VistACount(), # SHOULD BE IN OWN GRAPHIC
    
        "SSN": cd.idPresence(CommonUserDirectory.ENTRY_SSN_IDX),
        "SECID": cd.idPresence(CommonUserDirectory.ENTRY_SECID_IDX),
        "NWNM": cd.idPresence(CommonUserDirectory.ENTRY_NWNM_IDX),
        "DUID": cd.idPresence(CommonUserDirectory.ENTRY_DUID_IDX)
    }
    plotData["userPercs"] = {
            "title": f'{countsFromTotal["__total"]} Users',
            "plotName": f'{PLOT_PREFIX}Percs',
            "plotMethod": "plotIsHasBH",
            "specs": countsFromTotal
    }
    blurb = "The directory holds <span class='yellowIt'>{:,}</span> Users. The majority are in greater than one VistA. Most have a social security number (SSN) while a much smaller number have VA's own \"security id\" (secid).".format(countsFromTotal["__total"])
    mu += "{}\n\n{}\n\n".format(
        blurb, 
        muPlotRef(plotData["userPercs"], imageDir)
    )
     
    commonTotals = defaultdict(dict) # not picturing           
    allCnt = len(cd.vistas())
    for vistaInfo in cd.vistas():
        stationNo = vistaInfo["stationNumber"]
        stationMN = vistaNames[stationNo]
        commonCountsNos = cd.commonCounts(stationNo)
        commonCountsPer = {
            "all": commonCountsNos[allCnt],
            "some": sum(commonCountsNos[i] for i in range(2, allCnt)),
            "exclusive": commonCountsNos[1]
        }
        commonTotals[stationMN] = commonCountsPer
        
    plotData["commonByVistA"] = {
            "title": "Common User Count",
            "plotName": f'{PLOT_PREFIX}Common',
            "plotMethod": "plotCategoryBSV",
            "specs": commonTotals
    }
    blurb = "The graph below shows the total number of users in each system as well as a breakdown of whether those users are unique to that system (exclusive), in some but not all systems (some) or in every (all) other systems in the directory. It's striking how few of the users in the smaller systems - Walla Walla (WWW) and Spokane (SPO) - are unique to them. Indeed, for WWW, most of its users appear in some or every other system."
    mu += "{}\n\n{}\n\n".format(
        blurb,
        muPlotRef(plotData["commonByVistA"], imageDir)
    )
    
    plotData["commonByVistAPerc"] = {
            "title": "Common User Count (Percentage)",
            "plotName": f'{PLOT_PREFIX}PercCommon',
            "plotMethod": "plotCategoryBSV",
            "specs": commonTotals,
            "kargs": { 
                "usePerc": True
            }
    }
    blurb = "The following shows the same information but instead of totals, each system's representation distinguishes the percentage of its users that are unique from those that also appear elsewhere. Puget Sound, the biggest system, has the highest proportion of unique users. Clearly most users in smaller VistAs also use or have some impact on other VistAs."
    mu += "{}\n\n{}\n\n".format(
        blurb,
        muPlotRef(plotData["commonByVistAPerc"], imageDir)
    )
        
    return plotData, mu
    
# #################################### Time Series Section ##########################

"""
Stories shown User:
- last three full months across all four systems, 12/19-2/20, shows evenness, holiday and weekend dips and Puget adding more users than other systems. All 'marching along' inside Puget, Portland trying to reach up.
- pull back and look at yearly trends, there are four main changes to such evenness, in 
- last five years, monthly means highlights how different 2019 was. Portland's additions exceed Pugets and both these VistAs and Spokane add far more users than normal. But note (and TODO: follow up), the mid peek of POR and SPO seems to pre-existing users.
- last five years annual totals show how different 2019 was, both in volume for all systems except Walla Walla and, for once, Portland adds more users than Puget
"""

def makeTimeSeriesSection(cd, vistaNames, imageDir="Images"):

    plotData = {}
    
    end = "2020-02" # so common for all (make dynamic going forward)

    mu = "## User Additions over Time\n\n"
    
    dailiesDFCSV = csvDatesBySNO(cd)
        
    colNameById = vistaNames.copy()
    colNameById["USER"] = "New" # new users
    
    plotData["tsDailies"] = {
        "title": "2020/01-{} User Addition".format(re.sub(r'\-', '/', end)),
        "plotName": f'{PLOT_PREFIX}',
        "plotMethod": "plotDailies",
        "dfcsv": dailiesDFCSV,
        "kargs": {
            "colNameById": colNameById,
            "start": "2020-01",
            "end": end
        }
    }
    blurb = "The following shows __Daily User Addition__ for the last full three months for which data is available with Mondays marked on the X-axis. As expected addition mainly happens on weekdays with holidays (1/1, 1/20 and 2/17) and weekends showing little activity. December shows less activity than the following months with a steady decline through the end of the month. As expected, Puget (PUG) adds the most users but given the disparity in system size - something that shows up in the Patient Directory Report - it is surprising how close the numbers for all systems are. _This needs examination_."
    mu += "{}\n\n{}\n\n".format(
        blurb, 
        muPlotRef(
            plotData["tsDailies"],
            imageDir
        )
    )
        
    plotData["ts365RollingMean"] = {
        "title": "Trends in User Addition (365 day rolling)",
        "plotName": f'{PLOT_PREFIX}',
        "plotMethod": "plot365RollingMean",
        "dfcsv": dailiesDFCSV,
        "kargs": {
            "colNameById": colNameById,
            "start": "1987-01", 
            "end": end
        }
    }
    blurb = "Here are the yearly trends using a __\"365 day rolling mean\"__ starting when the directory first gets data to the latest additions. In general, additions follow system size but there's an exception in 2019 where four of the seven have a large spike in additions and Portland exceeds Puget's total. This is one of four times where the rate of addition changes - the others are 2002, 2006 and 2009. _All need examination_ - they must reflect the addition of or changes to VistA applications and use."
    mu += "{}\n\n{}\n\n".format(
        blurb,
        muPlotRef(
            plotData["ts365RollingMean"],
            imageDir
        )
    )
    
    plotData["tsMonthlyResamples"] = {
        "title": "Users - Resampled Monthly (2018 - {})".format(re.sub(r'\-', '/', end)),
        "plotName": f'{PLOT_PREFIX}',
        "plotMethod": "plotMonthlyResamples",
        "dfcsv": dailiesDFCSV,
        "kargs": {
            "colNameById": colNameById,
            "start": '2018-01', 
            "end": end
        }
    }
    blurb = "When the data is __resampled by month__, it shows three subpeaks for the 2019 leap in additions. The reason for Portland's leap above Puget also shows more clearly - though Puget has one large spike in April 2019, Portland has two, one in July and one in September. Note that the lack of any spike mirrors the flat line for smaller systems during this period in the yearly trends above.\n\nThis graph also shows (\"New\") the running total of new users added. As seen above, users added to VistAs sometimes already pre-exist in other VistAs. While the new user line shadows Puget's peak in April and Portland's in September, it is markedly lower than the spikes for Portland and Spokane in July - this indicates that much of those spikes are existing (Puget?) users being added to these systems. _This needs examination_."
    mu += "{}\n\n{}\n\n".format(
        blurb,
        muPlotRef(
            plotData["tsMonthlyResamples"],
            imageDir
        )
    )
  
    plotData["tsTotalsPerYear"] = {
        "title": "Users - Annual Additions (2015 - {})".format(re.sub(r'\-', '/', end)),
        "plotName": f'{PLOT_PREFIX}',
        "plotMethod": "plotTotalsPerYear",
        "dfcsv": dailiesDFCSV,
        "kargs": {
            "colNameById": colNameById,
            "start": "2015", 
            "end": end
        }
    }
    # TODO: calc these #'s - dangerous to just put here
    blurb = "Finally, a simple __side by side set of totals__ for the last five years echo's the graphs above with Portland jumping ahead of Puget in 2019. That year, Portland added <span class='yellowIt'>60,000</span> users, twice the number of a \"normal\" year. Spokane, too nearly doubled its numbers while Walla Walla stayed steady with under <span class='yellowIt'>20,000</span> additions. The years prior to 2019 share the same steady pattern of addition."
    mu += "{}\n\n{}\n\n".format(
        blurb,
        muPlotRef(
            plotData["tsTotalsPerYear"],
            imageDir
        )
    )
                                    
    return plotData, mu
    
# ######################### Utils used by Patient Too ######################
    
"""
DF Start
(will move to own utils)

A FOLLOW DIRECTLY: https://medium.com/li-ting-liao-tiffany/solving-real-world-business-questions-with-pandas-70ef8ef02675
... see all the techniques and trade offs 
ie/ cleaning, grouping etc and then compare to Typer

Basic DF to Typer:
- col per subTypeId (fixed value) + one other col with variable values
  ie/ <=> groupby?
  ... many individually grouped by
=> N DF's based on # of props + extra for create as "dayOfWeekCreated" etc.
(do first on Placer Consults -- see notes)

DF for all data exercise (perhaps enum -> C: only, timestamp down?, )
- Basic DF exercise --- how big would a straight up loading of rows get? (count of mults)
  ... load all
  ... ts.memory_usage(deep=True)
  ... idea of "store as a categorical!" ie/ enum of values elsewhere and ref with int ... may be better ... ie/ often "object" -> "category" ts2['name'].astype('category')
see: https://pandas.pydata.org/pandas-docs/stable/user_guide/scale.html

DF default reports and plotting <----------- KEY
  
Typer Improve Goals too:
... rem: context is byType grouping that doesn't loose granularity ie/ it's already
a pre-grouping. When going to pandas, gotta work out what I want to see in pandas.
ex/ if all by service of IFC => one service value and could see range of that etc ...
- byValueCount --- just put straight ie/ {}
- use __ for all meta including __count etc
- date inline too and all others?
- byHour option to go with byDay, byMonth, byYear + Day of Week (sample rate)
- option to reference off to labels for pters => if used in subtypes over all over!
      ... 200 ref efficiency ie/ label out ... may do just for it for now (and can expand those types) ... ie/ only 200 at start as others could balloon ... then do others if # lvalues << # asserts ... do comparison 
        ... will need big code change! ie/ do see 200- and then ...
        ... also need a convert new to old type if need be ... [util so no need to regen all]
... _ _count ala _subTypeId
- want the setup options (date time granularty etc to be obvious up front (and drive pandas?) and have a separate .info option?
- import to pandas option (and can see pandas data usage using nvy_df.info)

Takes result of directory.dumpTimeSeriesData() into a DataFrame

https://medium.com/sfu-cspmp/advanced-visualization-for-data-scientists-with-matplotlib-15c28863c41c (see techniques)

FIRST:
======
ie/ df aggregation into a column ala when I combine sts ... if I do a column per criteria and have their vals (subtypeid) and then row each for values 
... then can recombine

> df = df[['PID', 'YEAR_BUILT']].groupby('YEAR_BUILT', as_index = False).count().astype('int').rename(columns = {'PID':'No_of_properties_built'})

and

> df = df[(df['YEAR_BUILT'] >= 1900) & (df['YEAR_BUILT'] <= 2018)]

Look into https://stackoverflow.com/questions/34193862/pandas-pivot-table-list-of-aggfunc
with aggfunc for stats and the pivottable (explained!)

SECOND:
=======
- typer to df 
https://medium.com/well-red/cleaning-a-messy-dataset-using-python-7d7ab0bf199b
... along with subtype id to rows as flatten into rows ... also will drop optionals? as typing not strict OR must populate a default or "N:NONE" for catags
- use other cleans like normalize catag names USA, united states unify ... ie/ df.py (do my own df utils (mainly tips til get routines)

Could have used JSON ...

df = pd.DataFrame({'Date': [day1, day2 ...],
                   '668': [cnt day 1, cnt day2, ...],
                   '663' ...
                   })

or DataFrame({...}, [dates as index])                   

vs CSV form

    Date,668,663,648,687
    1979-02-13,0,0,1,0
    ...
    
https://hackingandslacking.com/plotting-data-with-seaborn-and-pandas-d2499fdf6f01 <---- good on "cleanup CSV pandas" and objects [if doing own typing]
    
https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.to_markdown.html includes markdown serialization

TODO: 
- https://towardsdatascience.com/pyviz-simplifying-the-data-visualisation-process-in-python-1b6d2cb728f1 <------ pandas built in "visuals" ... max/min etc. ie/ just working the
dataframes before going to deeper graphs
- https://towardsdatascience.com/handling-pandas-groupby-and-its-multi-indexes-efae3e6b788c with various grouping etc techniques
- https://tryolabs.com/blog/2017/03/16/pandas-seaborn-a-guide-to-handle-visualize-data-elegantly/ ... see groups columns etc

Other options to look at: (https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_csv.html)
- if no names are passed the behavior is identical to header=0 and column names are inferred from the first line of the file
- squeeze: If the parsed data only contains one column then return a Series. (TODO: using a series)

Note: if want CSV inline then use StringIO
    pd.read_csv(io.StringIO(data))
where data is string CSV

Can just make a DF with explicit Index (column) and other columns ...

speed = [0.1, 17.5, 40, 48, 52, 69, 88]
lifespan = [2, 8, 70, 1.5, 25, 12, 28]
index = ['snail', 'pig', 'elephant',
         'rabbit', 'giraffe', 'coyote', 'horse']
df = pd.DataFrame({'speed': speed,
                   'lifespan': lifespan}, index=index)
                   
In Dailies:
    index = dates (col 0 of csv)
    data = {"COLNAME": [...] ie/ "668": [...] etc.
    where COLNAME appears in legend

MORE:
- https://levelup.gitconnected.com/pandas-dataframe-python-10-useful-tricks-b4beae91df3d ... filtering, combining, renaming in frames ie/ all frame operations to try
- https://medium.com/@rrfd/cleaning-and-prepping-data-with-python-for-data-science-best-practices-and-helpful-packages-af1edfbe2a3 <---- includes 'fillna'
"""
def csvDatesBySNO(cd=None, datesBySNO=None):
        if datesBySNO == None:
            datesBySNO = cd.entryDatesBySNO()
        snosOrdered = sorted(datesBySNO, key=lambda x: len(datesBySNO[x]), reverse=True)
        byDateBySNO = defaultdict(Counter)
        for sno in datesBySNO:
            for dt in datesBySNO[sno]:
                byDateBySNO[dt][sno] += 1
        csvLines = []
        for dt in sorted(byDateBySNO):
            ln = [str(byDateBySNO[dt][sno]) if sno in byDateBySNO[dt] else "0" for sno in snosOrdered]
            ln.insert(0, dt)
            csvLine = ",".join(ln)
            csvLines.append(csvLine)
        cols = snosOrdered
        cols.insert(0, "Date")
        csvLines.insert(0, ",".join(cols))
        return csvLines
                    
# ############################## DRIVER ################################
            
def main():

    assert sys.version_info >= (3, 6)
    
    if len(sys.argv) > 1:
        if sys.argv[1] == "PLOT":
            makePlots("Common", "Users")
            return
        raise Exception("Only argument allowed is 'PLOT'")

    visn20VistAs = vistasOfVISNByOne("20")

    dirName = "User"
    cud = CommonUserDirectory(visn20VistAs)

    plotData = {}
    mu = f"""---
layout: default
title: VISN 20 {dirName} Directory
---

# VISN 20 {dirName} Directory

"""

    mu += """The following reports on a VISN 20 User Directory built from the user file (200) of 7 VistAs: Puget Sound [663], Portland [648], Boise [531], Spokane [668], Roseburg [653], White City [692] and Walla Walla [687]. Together they hold <span class='yellowIt'>{:,}</span> distinct users using <span class='yellowIt'>{:,}</span> user file records.

Note: Anchorage is the missing VISN 20 system. Its information will be added once the system is made available.

""".format(
        cud.total(),
        cud.entryTotal()
    )
    
    ttlPlotData, ttlMU = makeTotalsSection(cud, visn20VistAs)
    mu += ttlMU
    plotData.update(ttlPlotData)
    
    tsPlotData, tsMU = makeTimeSeriesSection(cud, visn20VistAs)
    mu += tsMU
    plotData.update(tsPlotData)
        
    plotDir = f'{VISTA_DATA_BASE_DIR}Common/TmpWorking/'
    print(f'Serializing vizData to {plotDir}')
    json.dump(plotData, open(f'{plotDir}plotDataUsers.json', "w"), indent=4)
                
    siteDir = SITE_DIR_TEMPL.format("Common")
    print(f'Serializing Report to {siteDir}')
    open(f'{siteDir}visn20UserDirectory.md', "w").write(mu) 
        
if __name__ == "__main__":
    main()
    


#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict, Counter
from datetime import datetime

from fmqlutils.reporter.reportUtils import MarkdownTable, reportPercent, reportAbsAndPercent
from fmqlutils.typer.reduceTypeUtils import splitTypeDatas, checkDataPresent, singleValue
from ..webReportUtils import ensureWebReportLocations, SITE_DIR_TEMPL, loadSelectTypes, reduce200, mdFilesOfDomain

"""
Dental with tie ins to Documents, Images, User (permissions) and Workload

Requires:
- 220_5 (Reduction) DENTAL PROVIDER
- 228_1 (SOReduction) DENTAL HISTORY ("dental encounter")
- 8925 (SOReduction)
- SELECT TYPEs for Dental File Types

TODO v1.1: notes and RPCs
- blurb ala PRF and others ie/ start with blurb
- Add 221: note that 221 is OLD ie/ last date of it? (being downloaded)
  > This file (228_1) is a new dental activity file. It is not meant to replace the official DAS at this time. Data stored in this file will be different from that in the DAS file (221)
- classes for dental notes?
- see others who use dental notes ie/ beyond providers identified
  > The dentists can do their Progress Note in the DSS-GUI which will download it to the VistA TIU package
- dental RPCs (key tie in to User Permissions) ... for DSS-GUI

TODO v2: introduce graph of activity per month + locns for Workload/PCE
- 228_1 SO ... show # per month over year, may break by type and consider per User breakdown (ala other types)
- 228_1 ... see locations (would tie to workload?): schema says locn for PCE!
  > at the same time, it will create PCE workload entries (ie/ auto created when putting in 228_1 with DSS-GUI

TODO v3: add to dental encounter ie/ activity file with this nuance
- Add 228_2: 
  - note always (100%) refers to 221 (less of them) => 221's group ie/ nuance
    ... pointer of 'dental encounter' ... may rename and use this phrase for 228_1 below?
  - Can use 228.3 to categorize? Most "type of care" is just transaction; visit_date break? 

TODO v4:
- tie in to imaging

ie/ doc, imaging, workload tie in and all dental done by end

TODO: see other xxrepo notes <---- nix once crossed
"""

"""
TODO MORE NUANCE: Orders/TIU

# From V PROVIDER of DENTAL Visits

vast majority of Dental Visits have one provider with

  PERSON CLASS: Dental Providers
  P:PRIMARY
  DATA SOURCE: DENTV DSS GUI, PACKAGE: DENTAL

but few are

  DATA SOURCE: TIU ..., PACKAGE: ORDER ...

==> came from orders but orders for what?
"""

def webReportDental(stationNo):

    ensureWebReportLocations(stationNo)

    allThere, details = checkDataPresent(stationNo, [
        {"fileType": "220_5", "check": "TYPE"}, # want Dental Providers 
        {"fileType": "228_1", "check": "TYPESO"}, # no split
        {"fileType": "8925", "check": "TYPESO"},
        {"fileType": "200", "check": "ALL"} # Index
    ])
    if not allThere:
        raise Exception("Some required data is missing - {}".format(details))

    mu = """---
layout: default
title: {} Dental
---

## Dental Package

""".format(stationNo) 

    # Practice of Care    
    type220_5, subs = splitTypeDatas(stationNo, "220_5", expectSubTypeProperty="")
    userInfoByUserRef = reduce200(stationNo)
    allDentalProviders = set(userRef for userRef in type220_5["name"]["byValueCount"])
    # Forcing use of SO signon ie/ last year only
    type3_081, st3_081 = splitTypeDatas(stationNo, "3_081", reductionLabel="YR1", expectSubTypeProperty="user")
        
    allUserSignOnCounts = dict((singleValue(st, "user"), st["user"]["count"]) for st in st3_081)
    currentDentalProviders = set(userRef for userRef in allDentalProviders if userRef in allUserSignOnCounts)
    # Dental History (228.1) == Dental Activity
    typeSO228_1, subs = splitTypeDatas(stationNo, "228_1", reductionLabel="YR1", expectSubTypeProperty="")
    allDentalRecordCreators = set(typeSO228_1["creator"]["byValueCount"])
    if len(allDentalRecordCreators - currentDentalProviders):
        raise Exception("Expected all dental record creators to be Current Dental Providers")
    
    # Note: this system doesn't use 220_2 in 220_5 ... dental classification -- so
    # fall back on titles.
    # ... TODO: more on Dental Activity Records (and related 228.2) for type of activity
    if "byValueCount" not in typeSO228_1["creator"]:
        raise Exception("Expect full byValueCount for {}".format("creator"))
    if "byValueCount" not in typeSO228_1["patient"]:
        raise Exception("Expect full byValueCount for {}".format("patient"))
    mu += """The system identifies <span class="yellowIt">{:,}</span> Dental Providers but only <span class='yellowIt'>{}</span> are current users, meaning they signed on to the system between _{}_ and _{}_, the most recent year for which data is available. There are <span class="yellowIt">{:,}</span> Dental Activity Records (\"Dental History\") for this period for <span class="yellowIt">{:,}</span> distinct patients treated by <span class="yellowIt">{:,}</span> of the dental providers (\"Dental Activity Providers\"). <span class="yellowIt">{:,}</span> were outpatient records, <span class="yellowIt">{:,}</span> in-patient.
    
""".format(
        len(allDentalProviders),
        reportAbsAndPercent(len(currentDentalProviders), len(allDentalProviders)),
        type3_081["user"]["firstCreateDate"].split("T")[0],
        type3_081["user"]["lastCreateDate"].split("T")[0],

        typeSO228_1["_total"],
        len(typeSO228_1["patient"]["byValueCount"]),
        len(typeSO228_1["creator"]["byValueCount"]),

        typeSO228_1["patient_type"]["byValueCount"]["O:outpatient"],
        typeSO228_1["patient_type"]["byValueCount"]["I:inpatient"]
    )
    
    # Now let's do a document break down [1] record providers, [2] others
    all8925, subTypes8925 = splitTypeDatas(stationNo, "8925", reductionLabel="YR1", expectSubTypeProperty="document_type")
    countByDocTypeByDentalProvider = defaultdict(lambda: Counter())
    parentTypeByDocumentType = {}
    for subTypeInfo in subTypes8925:
        if re.search(r'(8925_1\-0|ADDENDUM|ERRONEOUS)', subTypeInfo["_subTypeId"]):
            continue
        matched = False
        docType = singleValue(subTypeInfo, "document_type")
        parentDocType = singleValue(subTypeInfo, "parent_document_type")
        for pProp in ["expected_signer", "entered_by"]:
            if pProp not in subTypeInfo: # one in 668 (so rare!)
                continue
            pPropVC = subTypeInfo[pProp]["byValueCount"]
            matchedUserRefs = set(userRef for userRef in pPropVC if userRef in allDentalProviders)
            if len(matchedUserRefs):
                for userRef in matchedUserRefs:
                    countByDocTypeByDentalProvider[userRef][docType] = pPropVC[userRef]
                parentTypeByDocumentType[docType] = parentDocType
                break # only need one matcher - if not entered_by then expected_signer
   
    def tableDentalProviders(providers):
        count = 0
        tbl = MarkdownTable(["Provider", "Title", "Sign on \#", "Document \#s"])
        for userRef in sorted(countByDocTypeByDentalProvider, key=lambda x: sum(countByDocTypeByDentalProvider[x][y] for y in countByDocTypeByDentalProvider[x]), reverse=True):
            if userRef not in providers:
                continue # for other providers which may not write documents
            count += 1
            dtDetailsMU = ", ".join(["{} ({:,})".format(docType.split(" [")[0], countByDocTypeByDentalProvider[userRef][docType]) for docType in sorted(countByDocTypeByDentalProvider[userRef], key=lambda x: countByDocTypeByDentalProvider[userRef][x], reverse=True)])
            tbl.addRow([ 
                "__{}__".format(userRef.split(" [")[0]), 
                userInfoByUserRef[userRef]["title"] if "title" in userInfoByUserRef[userRef] else "",
                allUserSignOnCounts[userRef],
                dtDetailsMU
            ])
        return tbl, count
                
    mu += """Dental Providers write documents (\"TIU Notes\"). The most significant are written by the <span class="yellowIt">{:,}</span> _Dental Activity Providers_, those who create _Dental Histories_ (228.1) ...
    
""".format(len(allDentalRecordCreators))
    tbl, count = tableDentalProviders(allDentalRecordCreators)
    mu += tbl.md() + "\n\n"  
    
    tbl, count = tableDentalProviders(currentDentalProviders - allDentalRecordCreators)
    mu += """The remaining <span class="yellowIt">{:,}</span> Dental Providers with documents are a mix of _MSAs_ and other assistants who enter acknowledgements and consents not all of which may be about dental services ...
    
""".format(count)
    mu += tbl.md() + "\n\n"
    
    usedFiles = ["DENTAL PROVIDER", "DENTAL HISTORY", "DENTAL PATIENT"]
    mu += mdFilesOfDomain(
        stationNo, 
        "Dental Package", 
        220, 
        230,
        usedFiles
    )
    
    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    open(userSiteDir + "dental.md", "w").write(mu)
        
# ################################# DRIVER #######################
               
def main():

    assert sys.version_info >= (3, 4)

    try:
        stationNo = sys.argv[1]
    except IndexError:
        raise SystemExit("Usage _EXE_ STATIONNO")

    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    if not os.path.isdir(userSiteDir):
        raise SystemExit("Expect User Site to already exist with its basic contents")
    
    webReportDental(stationNo)
    
if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict, Counter
from datetime import datetime, date, timedelta
from dateutil.relativedelta import relativedelta

from fmqlutils import VISTA_DATA_BASE_DIR
from fmqlutils.cacher.cacherUtils import FMQLReplyStore, FilteredResultIterator, metaOfVistA
from fmqlutils.reporter.reportUtils import MarkdownTable, reportPercent, reportAbsAndPercent, muBVC
from fmqlutils.typer.reduceTypeUtils import splitTypeDatas, checkDataPresent, singleValue, combineSubTypes, muBVCOfSTProp, refsOfST

from fmqlreports.webReportUtils import TOP_MD_TEMPL, SITE_DIR_TEMPL, ensureWebReportLocations, keyStats, flattenFrequencyDistribution, roundFloat, reduce4, flattenPropValues 
from fmqlreports.webReportUtils import muPlotRef, makePlots, vistasOfVISNByOne

"""
NEXT ON CONTENT (not on visuals):

- Filler breakdown ala Placer (no mixing for now)
- may only do completed ie/ nix discontinued/cancelled for now
- expose remote site service name for placers? ie/ to link em up?
    ifc_remote_service_name (pair with local services so see if 1 remote for > 1 local?)
- visn named (not just other visns)
- filler only patients (delay reg?) ... what?
- WILL START with COMPLETED vs OTHER for all CONSULTs and then just do COMPLETED in rest of report.
- locations:
    from/... location 44 (no pattern seen per see but do they all have similar workloads?)
    ordering_facility (is precise to clinic)

Portland vs Pug as big (can do TeleMental as part of filler breaks):
... TODO:
    - TeleMental highlight for POR from all four others?
    - setup BOI for Pharmacy the same
- portland has more filler only services
- portland has more placer consults
- pug has more filler only patients (because proportionately more FILLER IFCs ie/ few placers)
- their user proportions are remarkedly similar
- common placers (50% PUG, 76% POR)
- spec placers (20 PUG, 5 POR)
  - for PUG, POR dominates, for POR, BOISE dominates
  - PUG to POR is for TELEMENTAL/ TELE NEURO! WWW does this the most for POR too and SPO has it too [TELEMENTAL is POR's second highest FILLER after TELER ie/ POR specializes in this]
  - POR to BOI (smaller amount) is Pharmacy! [NEED TO SEE HOW MUCH PHARMACY and not just TELER that BOI does]
  - lessor spec placees differ:
    - CANCER TEAM from PUG (expected!) but not POR
  - few other VISNs ... no real pattern
  
FORMATTING IMPROVEMENTS <--- key for going forward
- 1. MOVE all calcs to "Data Prep/Frame Phase" ... also key stats (vs so many details) out
- 2. Consider plots separate ie/ can gen first and then write blurbs ... could cut down on data needed for actual text/paragraphs ... could do tables in this phase too ie/ table of DF vs Graph of DF ... plot from frames/table from frames
- 3. Better struct for md ... add section, add header, add paragraph, add table ...
"""

"""
Next week should be last on IFCs (for now): it'll add Puget and flesh out the findings for the other three VistAs (the intra and inter VISN aspects as well as the Filler as opposed to the Placers of IFCs).
            	+
Consider how to see range of test consults in B1930 [ie how to feed into that]
"""

""" 
TODO NEXT: for basic, do filler 
<----- goal is finish IFC in three pushes along with TeleR (into 2 of 3) ... deliver mid Aug
- travel: break to VISN out and shorten table
- filler:
  - VIRS, DISCHARGE FOLLOW UP, TRAVELLING VETERAN, OTHER (do simply for now to parallel)
  ... then try PUG ... and will flip FILLER/PLACER on Volume
- ? DISCHARGE follow up in FILLER -- what is it on the other end? as not obvious
  ... add note TODO now
- OTHER Q: if high # local -- is there a type of things ALWAYS DONE LOCALLY?
- go to DF (before doing services more break etc) ... typerDFPlay
- break out other services
  - one reason for so many placer services is 'embeds destination' => for any consult type, need where as well (any exs of 'type for x', 'type for y'? ex/ ENDOCRINOLOGY (visit and non visit) ... POR and SEATTLE ... UROLOGY is another example ... with many many. see if can unify using formal meta? ie disciplines, smaller is PODIATRY but still over three centers
  <------- break out other services
  - Break out VIRs 
  - IVR is "intra VistA Req" (sic) ... what is it text-wise (123 look)
- do graph per weekday as have it for each (more/new Py code below)
  - redo reduction for per month (or per day?) to get time series
  <------ chance to use per weekday <------- gets into time series too
- get into Filer side for PUG etc ie/ WANT MORE ON FILER SIDE (may be week three of IFC?)
- more on if Consult has images or not (teler beyond on image bearing consults?)
- user and patient discrepancies ... a/c for them <---- make top bars relevant
- OTHER DATA: TIU, Image, Appts - any connection to consults (IFCs) ie/ anything else made? [will tie to images and encounters] ... also want to see if any have img or doc assoc? [123 has TIU link! and assoc results ie/ PROCESSED IN TELEREADING (do this in there first ie/ any image etc ref from doc?]
- VISN 20 dynamic! ie/ lookup for any system
- clean TODO PASS 2 below

USE AS EXCUSE FOR:
> images and encounters (We are trying to determine the encounter relationships with the images at the VA. Hoping the data you have could help us. Could you provide how many images have encounter relationships for the IOC Sites with the Cache database you have?
Justin Cook; Botzum,Ed <Ed.Botzum@cerner.com>: Also, could we get a sense of what percent of the total that is. I.E. 72% of all images are associated with encounters.)
         ... work back from Docs and Visits? or? <---- put into a report
"""
    
"""
All Consults - IFC Consults with TeleRead as Subset

Bonus: enforces that all routing is to or from primary station #'s/4's and
not (sub) clinics. ie/ are VistA and not subfacility stations and ids.

Relies on: prepareMeta for 4 (may go native on it when move)

TODO PASS 2:
- 123_5:
  - and its stop code ... bring stop code in here
  - Service Remote name (is in 123_5 too) for IFC placer ... ie/ these are my PLACER five ... probably worth doing [true for more than TeleReading?]
- BIG: 2005, 8925 tie ins + PATIENT timing (longest time gap? ... need enhanceResource 123)
  ... may be through 2006.84/5 in 2005 itself or from 123_5 (indexed in VistA Imaging)
  ... but is the link "manual"? how link to document? (may need extra q)
- Locations (Clinics) ie/ precise ones
- More on Users and Roles and particularly default (programmed) users (fill accepted?)
- COMPLETE has result always as essential? Why not -- go back to mand/opt

TO WORK FILLER SEPARATION:
- PORTLAND
    DIABETIC TELERETINAL IMAGING (PORTLAND) IFC - MANN-GRANDSTAFF VAMC [689], ROSEBURG VA MEDICAL CENTER [372], ANCHORAGE VA MEDICAL CENTER [326], BOISE VA MEDICAL CENTER [279], JONATHAN M. WAINWRIGHT VAMC [108], WHITE CITY VA MEDICAL CENTER [35]
    ie/ see SPO, ROSE, ANC, BOISE, WWW, WHITE CITY
  [in simil now, would only get SPO and WWW]
  [NOTE: 60% of SPO go to PORT, others go to BOISE]
- PUGET 
    TELEDERMATOLOGY IFC (PUGET SOUND) - SOUTHERN ARIZONA VA HCS [2,886], ROSEBURG VA MEDICAL CENTER [2,641], PHOENIX VAMC [2,108], BILOXI VA MEDICAL CENTER [1,746], JONATHAN M. WAINWRIGHT VAMC [1,194], MANN-GRANDSTAFF VAMC [1,192], WHITE CITY VA MEDICAL CENTER [545], BOISE VA MEDICAL CENTER [284], VA HEARTLAND - WEST, VISN 15 [188], ANCHORAGE VA MEDICAL CENTER [55], PORTLAND VA MEDICAL CENTER [2]
    DIABETIC TELERETINAL IMAGING IFC (PUGET SOUND) - BOISE VA MEDICAL CENTER [3], MANN-GRANDSTAFF VAMC [1] (prob no longer used - and see in SPO ... do date stamp to remove?)
- [Missing BOISE end of 40% SPO EYE]

<----------- [1] come back here and to [2] 2005 rep for below (XRAY GROUP == group of any etc) before going back + [3] We're going to add the edge case to our simulation data: spec/proc to two readers and see that both see work for the other. Mike saw that the TeleReader displays the "unintended" in a particular way. More on that next week.
"""
def webReportIFC(stationNo, stationName, onAndAfterDay, upToDay, imageDir="Images", doDAD=False):

    type123YR1, sts123YR1All = splitTypeDatas(stationNo, "123", reductionLabel="YR1", expectSubTypeProperties=["cprs_status", "routing_facility", "to_service"])
    sts123YR1 = []
    for st in sts123YR1All:
        try:
            singleValue(st, "to_service")
        except:
            print("** No service for {} ({:,}) so skipping".format(st["_subTypeId"], st["_total"])) # one in POR
            continue
        sts123YR1.append(st)
                
    institByIEN = reduce4(stationNo)
            
    # WRF - with routing facility => IFC ie/ wrf <=> ifc
    stWRF = [st for st in sts123YR1 if "routing_facility" in st]
    placerServicesCheck = set()
    # If routing_facility => ifc_role and vica_versa
    if sum(1 for st in stWRF if "ifc_role" not in st and "routing_facility" in st) or sum(1 for st in sts123YR1 if "ifc_role" in st and "routing_facility" not in st):
        raise Exception("Where routing facility expect role and where role, expect routing facility")
    try: # either placer or filler within st gathered by service, routing_facility
        for st in stWRF:
            ifcRole = singleValue(st, "ifc_role")
            # bonus - record placer service for below
            if ifcRole == "P:PLACER":
                placerServicesCheck.add(singleValue(st, "to_service"))
    except:
        print(json.dumps(st, indent=4))
        raise Exception("> 1 ifc_role in routing_facility'ed ST. Unexpected")
    byServiceWRF = defaultdict(list)
    for st in stWRF: # note: for filler/local overlaps will only gather the filler st's
        service = singleValue(st, "to_service")
        byServiceWRF[service].append(st)

    # Service overlap with RF may happen due to filler being used in RF 
    # and WORF consults
    #
    # REVISIT: handles case where Placer service DOES not get reused in WORF
    # (enforced below) BUT Portland has DISCONTINUED case where this happens.
    # ... perhaps discontinued happens before IFC info added to a Consult? 
    # ... TODO: follow up as not counting this now!
    # 
    stWORF = [st for st in sts123YR1 if not ("routing_facility" in st or singleValue(st, "to_service") in placerServicesCheck)]  
    # Just for tracing -- the WORF Placers
    # ... PENDING is fine (but rare); DISCONTINUED seems to be too and then WCO has
    # an Active that has no further action beyond CPRS submit and just sits there.
    stWORFPlacer = [st for st in sts123YR1 if "routing_facility" not in st and singleValue(st, "to_service") in placerServicesCheck]
    if len(stWORFPlacer):
        for st in stWORFPlacer: # POR (DISCONTINUED), PUG (PENDING)
            stat = singleValue(st, "cprs_status")
            if not re.search(r'(ACTIVE|DISCONTINUED|PENDING)', stat):
                print(json.dumps(st, indent=4))
                raise Exception("If a Placer Service ST lacks Route Info, expect it to be (quickly?) discontinued or to be pending")
        print("** Warning: {:,} consults were routeless to Placer Services.".format(
            sum(st["_total"] for st in stWORFPlacer)
        ))
    worfTTL = sum(st["_total"] for st in stWORF)
    worfServices = set(singleValue(st, "to_service") for st in stWORF)
    worfUsers = set(user for st in stWORF for user in refsOfST(st, "200"))
    worfPatients = set(patient for st in stWORF for patient in refsOfST(st, "2"))
    
    # if WORF removed a placer, must a/c for that so only doing ALL here.
    stAll = stWORF + stWRF 
    allTTL = sum(st["_total"] for st in stAll)
    allServices = set(singleValue(st, "to_service") for st in stAll)
    allUsers = set()
    allPatients = set()
    for st in stAll:
        allUsers |= refsOfST(st, "200")
        allPatients |= refsOfST(st, "2")
                
    # Big: one role per service ie/ placer or filer in practice
    wrfTTL = 0
    wrfUsers = set()
    wrfPatients = set()
    placerTTL = 0
    placerServices = set()
    placerUsers = set()
    placerPatients = set()
    fillerTTL = 0
    fillerServices = set()
    fillerUsers = set()
    fillerPatients = set()
    for service in byServiceWRF:
        serviceUsers = set()
        servicePatients = set()
        serviceTTL = 0
        for st in byServiceWRF[service]:
            serviceTTL += st["_total"]
            serviceUsers |= refsOfST(st, "200")
            servicePatients |= refsOfST(st, "2")
            # Enforce that a routing facility is PRIMARY ie/ placed to or
            # filled from a first class, VistA naming station
            rf = singleValue(st, "routing_facility")
            rfIEN = re.search(r'4\-([^\]]+)', rf).group(1)
            if rfIEN not in institByIEN:
                raise Exception("RF is for 4 entry with no Station No")
            if not re.match(r'\d{3}$', institByIEN[rfIEN]["station_number"]):
                print("{} - {}".format(rf, institByIEN[rfIEN]["station_number"]))
                raise Exception("RF points to 4 with non primary station number")
        wrfTTL += serviceTTL
        wrfUsers |= serviceUsers
        wrfPatients |= servicePatients
        roles = set(role for st in byServiceWRF[service] for role in st["ifc_role"]["byValueCount"])
        if len(roles) != 1:
            raise Exception("Only expect one role for a service ie/ placer or filer")
        if list(roles)[0] == "F:FILLER":
            fillerServices.add(service)
            fillerUsers |= serviceUsers
            fillerPatients |= servicePatients
            fillerTTL += serviceTTL
        else:
            placerServices.add(service)
            placerUsers |= serviceUsers
            placerPatients |= servicePatients
            placerTTL += serviceTTL
            if len(set(rf for st in byServiceWRF[service] for rf in st["routing_facility"]["byValueCount"])) > 1:
                # ie/ Service --> Place Routed to if Placer. Filler may take from
                # multiple places.
                raise Exception("Expected one and only one RF for Placer Services")
                
    if placerServices != placerServicesCheck:
        print(len(placerServices))
        print(len(placerServicesCheck))
        raise Exception("Placers analyzed before to avoid WORF problems above - make sure in Sync") # TMP til tackle the DISCONTINUED
    if (placerTTL + fillerTTL + worfTTL) != allTTL:
        raise Exception("Totalling mistake - consult totals are exclusive")
    # ONLY for no overlap placer and WORF ... filler may be reused so there can be
    # overlap (ie/ remote VistA reuses a locally used consult service)
    if len(worfServices.intersection(placerServices)):
        print(worfServices.intersection(set(placerServices)))
        raise Exception("Expect Placer to be exclusive from Local Services")
    fillerAndLocalServices = worfServices.intersection(fillerServices)
    if len(allServices) != len(placerServices) + (len(fillerServices) - len(fillerAndLocalServices)) + len(worfServices): # overlap only counted in worfServices
        raise Exception("Expected overlap of filler and local to be easy to a/c for in allTotal") 
        
    # Tele SubSet definitively leveraging 2006_5849 USE!    
    type2006_5849YR1E, sts2006_5849YR1E = splitTypeDatas(stationNo, "2006_5849", reductionLabel="YR1E", expectSubTypeProperties=["reader_duz_at_acquisition_site", "image_index_for_specialty", "consult_service", "status", "#reading_start"])
    # DFNOTE: would be derived column (ie/ placer service type: TELER | TRAVEL | OTHER)
    telerAcServicesUsed = set(singleValue(st, "consult_service") for st in sts2006_5849YR1E)
    
    # DFNOTE: other to break further -- NEUROSURGERY X, (gather all) -- see if any subtyping done formally?
    # ... DOMICIALLARY REFERRAL (https://www.visn2.va.gov/VISN2/bh/inpatient.asp)
    # ... TELEDERMATOLOGY ... why not in TELE... top one WWW to PUG
    # ... granularity of subcategory advised ... how to group so can assign?
    # ... do as part of typer upgrade for categories as types => limited # (can
    # include PTERs!) ... want reduction (proportion?)
                
    # Cache data on Image
    meta = metaOfVistA(stationNo)
    
    dad = DumpAndAnalyzeData(stationNo, meta["cutDate"]) if doDAD else None
                        
    # ######################## MU Consults Per ####################
    
    plotData = {}
         
    mu = TOP_MD_TEMPL.format("{} IFCs".format(stationNo))
         
    mu += "# Interfacility Consults (IFCs) {} [{}]\n\n".format(meta["name"], stationNo)
    
    mu += "The following describes Consults, specifically Interfacility Consults (IFC), for the most recent year of _{}_ VistA cut on {}.\n\n".format(meta["name"], meta["cutDate"])  
    
    # Blurb
    blurb = """_Local consults_ - those ordered and fulfilled within this VistA - represent fully <span class='yellowIt'>{}</span> of the consults created in this VistA during the last full year for which data is available. """.format(
        reportAbsAndPercent(
            worfTTL,
            allTTL
        )
    )
    if placerTTL > fillerTTL:
        blurb += """Of the IFCs, there are more _Placer_ consults (\"placed\" elsewhere) <span class='yellowIt'>{}</span> than _Filler_ consults (fulfilled by this VistA for another) <span class='yellowIt'>{}</span>. In terms of IFCs, this is a __Placer VistA__ ...
    
""".format(
        reportAbsAndPercent(
            placerTTL,
            allTTL,
        ),
        reportAbsAndPercent(
            fillerTTL,
            allTTL
        )
    )
    else:
        blurb += """Of the IFCs, there are more _Filler_ consults (fulfilled by this VistA for another) <span class='yellowIt'>{}</span> than _Placer_ consults (\"placed\" elsewhere) <span class='yellowIt'>{}</span>. In terms of IFCs, this is a __Filler VistA__ ...
    
""".format(
        reportAbsAndPercent(
            fillerTTL,
            type123YR1["_total"]
        ),
        reportAbsAndPercent(
            placerTTL,
            type123YR1["_total"]
        )
    )
        
    # Plot: Consults Vs Services
    plotData["allConsults"] = {
        "title": "All Consults",
        "plotName": "ifcAllConsults",
        "plotMethod": "plotCategoryBH",
        "rows": ["consults"],
        "columns": ["local", "placer", "filler"],
        "data": [
            (
                worfTTL, 
                placerTTL, 
                fillerTTL
            ), # consults
        ]
    }
    plotRef = muPlotRef(
        plotData["allConsults"],
        imageDir
    )
    
    # MU - Blurb + Plot: Consults Vs Services
    mu += "{}\n\n{}\n\n".format(
        blurb,
        plotRef
    )
    
    blurb = """VistA Consults are ordered using consult types called _services_. Local consults use <span class='yellowIt'>{}</span> of the services used in this system, less than their proportion of consults. In other words, more services are employed for ordering IFCs than their numbers would suggest. Notice also that some services used for local consults - <span class='yellowIt'>{}</span> - are also used for IFC fulfillment. While services used for placement are exclusively for IFCs, the same is not true for fulfillment.
    
""".format(
        reportAbsAndPercent(
            len(worfServices),
            len(allServices)
        ),
        reportAbsAndPercent(
            len(fillerAndLocalServices),
            len(allServices)
        )
    )
    
    plotData["allServices"] = {
        "title": "All Consult Services",
        "plotName": "ifcAllServices",
        "plotMethod": "plotCategoryBH",
        "rows": ["services"],
        "columns": ["local only", "placer only", "filler only", "local+filler"],
        "data": [
            (
                len(worfServices) - len(fillerAndLocalServices), 
                len(placerServices), 
                len(fillerServices) - len(fillerAndLocalServices),
                len(fillerAndLocalServices)
            ) # services 
        ]           
    }
    plotRef = muPlotRef(
        plotData["allServices"],
        imageDir
    )
    
    mu += "{}\n\n{}\n\n".format(
        blurb,
        plotRef
    )
    
    # This one differs between Placer and Filler VistAs
    plotData["patientsAndUsers"] = {
        "title": "Users and Patients",
        "plotName": "ifcPatientsAndUsers",
        "plotMethod": "plotCategoryBH",
        "rows": ["users", "patients"],
        "columns": ["local only", "placer only", "filler only", "mixed"],
        "data": [
            (
                len(worfUsers - wrfUsers), 
                len(placerUsers - (worfUsers | fillerUsers)), 
                len(fillerUsers - (worfUsers | placerUsers)), 
                len(worfUsers.intersection(wrfUsers))
            ),
            (
                len(worfPatients - wrfPatients), 
                len(placerPatients - (worfPatients | fillerPatients)), 
                len(fillerPatients - (worfPatients | placerPatients)), 
                len(worfPatients.intersection(wrfPatients))
            )
        ]            
    }
    plotRef = muPlotRef(
        plotData["patientsAndUsers"],
        imageDir
    )
    blurb = """As seen below, most patients with consults only have local consults and the majority of the balance have both local consults and IFCs (\"mixed\").
    
"""
    if placerTTL > fillerTTL:
        blurb += """A small minority of patients have only fulfilled consults. With fulfilled consults, a patient's home record is typically elsewhere so you would expect to see more such patients and less mixed patients. The lower number of such \"remote-only patients\" should be examined further. 

As for users, a smaller proportion are only involved in local consults. While no users only fulfill consults submitted from elsewhere, a relatively large set only place consults elsewhere. This set of \"placer only users\" needs further examination.

"""
    else:
        blurb += """As a larger _Filler VistA_, a large number of its patients only have filled IFCs - the patient's home VistA is elsewhere. 
        
As for users, most tend to only local consults or to a mix of local and filled consults.

"""       

    mu += "{}\n\n{}\n\n".format(
        blurb,
        plotRef,
    )
    
    # ################################ Placer ######################
    
    HOMEVISNVISTAS = vistasOfVISNByOne(stationNo)     
    
    placerServicesFacilities = Counter()
    telerTBL = MarkdownTable([":Placer Service", ":To Facility", "\#", "Status'"])
    placerServicesTeleR = set()
    travelTBL = MarkdownTable([":Placer Service", ":To Facility", "\#", "Status'"])
    placerServicesTravel = set()
    commonTBL = MarkdownTable([":Placer Service", ":To Facility", "\#", "Status'"])
    placerServicesCommon = set()
    spclzedTBL = MarkdownTable([":Placer Service", ":To Facility", "\#", "Status'"])
    placerServicesSpclzed = set()
    """
    Enforced here as grouping done here -- TODO: move grouping of placers up to data
    organization and not in mix of presentation.
    
    Of WRF's, only TELER has a local result (2 places for it - as a doc and in generic
    associated_results). But it is a disappointing reference (in SPO -- TODO: check
    more and wider by keeping all 8925 refs when typing) ala 
    
        >  "Please refer to Inter-facility Consult for results.\n\nA
utomatically generated note - signature not required.\n \n  Electronically Filed
: {DAY}\n                    by: {author-dictator}\n
 ",
    """
    def enforceOnlyTelerHasARnDoc(st):
        if "associated results" in st:
            raise Exception("WRF: TELER is only one expected to have local associated results")
        if "tiu_result_narrative" in st:
            raise Exception("WRF: TELER is only one expected to have local tiu - equivalent to associated resutls")
    for i, service in enumerate(sorted(placerServices, key=lambda x: sum(st["_total"] for st in byServiceWRF[x]), reverse=True), 1):
        tttl = sum(st["_total"] for st in byServiceWRF[service])
        rf = singleValue(byServiceWRF[service][0], "routing_facility")
        placerServicesFacilities[rf] += tttl
        if service in telerAcServicesUsed:
            groupLabel = "TELER"
            placerServicesTeleR.add(service)
            tblToUse = telerTBL
        elif re.search(r'(?i)travel', service):
            groupLabel = "TRAVEL"
            placerServicesTravel.add(service)
            tblToUse = travelTBL
        # follow-up is PUG, f/u is POR and these DISCHARGE FOLLOW UPs are only from big 
        # (POR/PUG) to small and not vica-versa
        elif re.search(r'(?i)(discharge f\/u|discharge follow\-up|virs |domiciliary|clinic f\/u|lodging)', service):
            groupLabel = "COMMON"
            placerServicesCommon.add(service)
            tblToUse = commonTBL
        else: 
            groupLabel = "SPECIALIZED"
            placerServicesSpclzed.add(service)
            tblToUse = spclzedTBL
        rfcntr = Counter()
        scntr = Counter()
        # Note: for Placer, only status varies (RF and by Defn, service is one)
        for st in byServiceWRF[service]:
            if groupLabel != "TELER":
                enforceOnlyTelerHasARnDoc(st)
            scntr[re.sub(r'100_01\-', '', singleValue(st, "cprs_status"))] += st["cprs_status"]["byValueCount"][singleValue(st, "cprs_status")]
            if dad:
                dad.dumpDocsOfST(groupLabel, st)
        serviceName = re.sub(r' +$', '', service.split(" [")[0]) 
        facMU = re.sub(r'4\-', '', rf)
        facMU = f'__{facMU}__' if re.search(r'\-(\d+)\]', rf).group(1) in HOMEVISNVISTAS else facMU 
        tblToUse.addRow([
            # "__{}__".format(serviceName) if tttl/placerTTL > 0.1 else serviceName,
            serviceName,
            facMU,
            tttl,
            muBVC(scntr)
        ])
        
    if dad:
        dad.finish()
        
    mu += "## Placer\n\n"
    
    def summarizePlacerFacilities(services, groupLabel):
        serviceFacilityCount = defaultdict(lambda: Counter())
        facilityCount = Counter()
        ssttl = 0
        for service in services:
            sttl = sum(st["_total"] for st in byServiceWRF[service])
            ssttl += sttl
            facility = singleValue(byServiceWRF[service][0], "routing_facility")
            serviceFacilityCount[facility][service.split(" [")[0]] = sttl
            facilityCount[facility] += sttl
        cols = [":Facility", ":Service", "Consults"]
        if sum(1 for facility in serviceFacilityCount if len(serviceFacilityCount[facility]) > 1): 
            cols = [":Facility", "Consults", ":Services"]
        tbl = MarkdownTable(cols)
        visnValues = []
        otherTTL = 0
        for facility in sorted(facilityCount, key=lambda x: facilityCount[x], reverse=True):
            facSNO = re.search(r'\-(\d+)\]', facility).group(1)
            if facSNO in HOMEVISNVISTAS:
                visnValues.append((facilityCount[facility], HOMEVISNVISTAS[facSNO]))
            else:
                otherTTL += facilityCount[facility]
            facMU = re.sub(r'4\-', '', facility)
            # serviceNameMU = "__{}__".format(serviceName) if facilityCount[facility]/ssttl > 0.01 else serviceName
            facMU = f'__{facMU}__' if facSNO in HOMEVISNVISTAS else facMU 
            if cols[1] == ":Service":
                row = [
                    facMU,
                    list(serviceFacilityCount[facility])[0],
                    reportAbsAndPercent(facilityCount[facility], ssttl)
                ]
            else:
                row = [
                    facMU,
                    reportAbsAndPercent(facilityCount[facility], ssttl),
                    muBVC(serviceFacilityCount[facility])
                ]
            tbl.addRow(row)
        sankeyVIZData = {
            "title": f'{groupLabel} Placers', 
            "plotName": f'ifc{re.sub(" ", "", groupLabel)}', 
            "plotMethod": "plotOutOtherSankey",
            "specs": {
                "outValues": visnValues,
                "toOtherTTL": otherTTL,
                "fromLabel": "SOURCE" 
            }
        }   
        plotData[f'placer{groupLabel}'] = sankeyVIZData    
        plotRef = muPlotRef(
            sankeyVIZData,
            imageDir
        )
        return tbl, set(facilityCount), ssttl, plotRef
        
    # Break into three.
    pbdmu = "" # placer breakdown mu
    tttlTeleR = tttlTravel = tttlCommon = tttlSpclzed = 0
    
    """
    TODO: may put these down with specialized ... ie/ after travel and common ie/
    part of specialized breakdowns and need to do others so IFCs covered in general.
    """
    if len(placerServicesTeleR):
        ftbl, facilities, tttlTeleR, plotRef = summarizePlacerFacilities(placerServicesTeleR, "TeleReading")
        pbdmu += "<span class='yellowIt'>{}</span> of the Placer Services are for __TeleReading__ to <span class='yellowIt'>{:,}</span> facilities ...\n\n".format(
            len(placerServicesTeleR), 
            len(facilities) 
        )
        pbdmu += plotRef + "\n\n"
        pbdmu += "Service by Service ...\n\n"
        pbdmu += telerTBL.md() + "\n\n"

    # For Travel, just use the facility table as 1-1 for most but PORT breaks that
    # ... also note broken from "common" as not just within VISN!
    # TODO: 
    # - consider change to have toVISNs ... ie/ show all except lowest numbers
    # including this VISN ... sunnier VISNs comment
    # - ... also travel coincides with summer/holidays?
    # - also who making these (docs or others?) 
    if len(placerServicesTravel):
        ftbl, facilities, tttlTravel, plotRef = summarizePlacerFacilities(placerServicesTravel, "Travel") 
        pbdmu += "<span class='yellowIt'>{}</span> of the Placer Services denote __travel__ to remote facilities and represent <span class='yellowIt'>{}</span> of placed consults. These prove to be the exception to the rule that a VistA places most IFCs within its own VISN.\n\n".format(
            len(placerServicesTravel), 
            reportPercent(tttlTravel, placerTTL)
        )
        pbdmu += plotRef + "\n\n"
        pbdmu += "By Facility ...\n\n"
        pbdmu += ftbl.md() + "\n\n"

    """
    A range of common "admin" (VIRS ...) and regular services (F/Us, lodging requests)
    sent between systems of any size, whether they are fillers or placers.
    
    Discharge F/U ; VIRS ; Domiciliary ; Clinic F/U; LODGING ... -i
    
    Ex/ VIRS Coordinator - IFC Spokane [53], Spokane Clinic F/U - IFC [48], Discharge F/U - Spokane IFC [6]
    
    The fillers (see below) of Smaller VistAs are dominated by such requests. For larger
    clinics, these common requests are proportionately less important.
    """
    if len(placerServicesCommon):
        ftbl, facilities, tttlCommon, plotRef = summarizePlacerFacilities(placerServicesCommon, "Common") 
        # All such consults stay within the VistA's VISN.
        # ADD: if Placer VistA -- Note that as the filler section below shows, services like these dominate the list of consults filled by a Placer VistA such as this one.
        pbdmu += """<span class='yellowIt'>{}</span> are \"Common\" Placer Services sent between both filler and placer VistAs. They represent a range of typical administrative (_\"VIRS\"_, _Lodging_, _Domiciliary_) and care requests to remote facilities (Follow ups). In this VistA, such requests represent <span class='yellowIt'>{}</span> of placed consults.
  
Note that equivalent services are filled by this VistA. Those are detailed in the _Filler Section_ below.      
        
""".format(
            len(placerServicesCommon), 
            reportPercent(tttlCommon, placerTTL)
        )
        pbdmu += plotRef + "\n\n"
        pbdmu += "By Facility ...\n\n"
        pbdmu += ftbl.md() + "\n\n"

    # TODO: Portland etc -- very very few compared to WWW ie/ here fillers take the lot
    # ... makes sense to group filler and placed specialized ... few if any for POR, all 
    # Fill!
    if len(placerServicesSpclzed):
        ftbl, facilities, tttlSpclzed, plotRef = summarizePlacerFacilities(placerServicesSpclzed, "Specialized")        
        # TODO: add later -- Placer VistA has many (relatively); a Filler has few - this is reversed in the Filler section    
        pbdmu += "The balance, <span class='yellowIt'>{}</span>, of the Placer Services to <span class='yellowIt'>{:,}</span> facilities, are mainly for specialized clinical services and represent <span class='yellowIt'>{}</span> of placed consults. Once again, the vast majority go to other VISN 20 facilities but the flow is mainly from smaller facilities to larger with the small placing with the large.\n\n".format(
            len(placerServicesSpclzed), 
            len(facilities), 
            reportPercent(tttlSpclzed, placerTTL)
        )
        pbdmu += plotRef + "\n\n"
        pbdmu += "By Facility ...\n\n"
        pbdmu += ftbl.md() + "\n\n"
    
    """
    TODO PLOT: move to a horizontal once more categories with % and totals and overall #
        teler -----------xx
        travel -xx
        common -------xx
        spec -------------------xxx 
    """
    plotData["placerConsultsNServices"] = {
        "title": "Placed Consults and Services",
        "plotName": "ifcConsultsNServicesPlaced",
        "plotMethod": "plotCategoryBH",
        "rows": ["consults", "services"],
        "columns": ["telereading", "travel", "common", "specialized"],
        "data": [
            (
                tttlTeleR,
                tttlTravel,
                tttlCommon,
                tttlSpclzed
            ),
            (
                len(placerServicesTeleR),
                len(placerServicesTravel),
                len(placerServicesCommon),
                len(placerServicesSpclzed)
            ), # placed services 
        ]           
    }
    plotRef = muPlotRef(
        plotData["placerConsultsNServices"],
        imageDir
    )
    
    # May break out 'VISN 20 CANCER CARE NAVIGATION TEAM IFC' -- another common, peer to
    # peer service (all do placer and filler)
    blurb = "There are <span class='yellowIt'>{:,}</span> __Placer Services__ - services that route/place consults __to__ <span class='yellowIt'>{:,}</span> other facilities. In VistA, a Placer Service identifies a specific remote (routing) facility which is always the main, default facility of a remote VistA. In this VistA, <span class='yellowIt'>{}</span> of the IFC Consults are \"Placed\" and the vast majority go __to other VISN 20 facilities__, facilities highlighted in bold in the tables below. Placer consults can be broken into four broad classes ...\n\n".format(
        len(placerServices), 
        len(placerServicesFacilities), 
        reportPercent(placerTTL, wrfTTL)
    )
    mu += "{}\n\n{}\n\n".format(
        blurb,
        plotRef
    )
    mu += pbdmu # placer breakdown mu
    
    # ############################# Filler ###########################
    
    """
    TODO: when break down
    - Break into TeleR/Travel (out of VISN!)/Common/Specialized
      ... or isn't TeleR just another form of specialized?
    - consider NOT counting ANY discontinued except at the top => ie/ in summary => smaller tables
    """

    tbl = MarkdownTable([":Filler Service", ":From Facilities", "\#", "Status'"])
    filledCntByPlacer = Counter()
    for i, service in enumerate(sorted(fillerServices, key=lambda x: sum(st["_total"] for st in byServiceWRF[x]), reverse=True), 1):
        tttl = sum(st["_total"] for st in byServiceWRF[service])
        rfcntr = Counter()
        scntr = Counter()
        for st in byServiceWRF[service]:
            placer = re.sub(r'4\-', '', singleValue(st, "routing_facility"))
            cnt = st["routing_facility"]["byValueCount"][singleValue(st, "routing_facility")]
            filledCntByPlacer[placer] += cnt
            rfcntr[placer] += cnt
            scntr[re.sub(r'100_01\-', '', singleValue(st, "cprs_status"))] += st["cprs_status"]["byValueCount"][singleValue(st, "cprs_status")]
        serviceName = re.sub(r' +$', '', service.split(" [")[0]) 
        # TODO: from facilities don't appear to be ordered
        tbl.addRow([
            serviceName,
            ", ".join(["{} [{:,}]".format(rf.split(" [")[0] if re.search(r'\[(\d+)\]', rf).group(1) not in HOMEVISNVISTAS else f'__{rf.split(" [")[0]}__', rfcntr[rf]) for rf in rfcntr]),
            reportAbsAndPercent(tttl, fillerTTL),
            muBVC(scntr)
        ])
    
    mu += "## Filler\n\n"
    mu += "There are <span class='yellowIt'>{:,}</span> __Filler Services__ - services that accept/fill consults __from__ other facilities. In VistA, a single Filler Service may be employed by more than one remote facility and may also be used for purely local consults. In this VistA, <span class='yellowIt'>{}</span> of the IFC consults are \"filled\", mostly from VistAs in the same VISN ...\n\n".format(len(fillerServices), reportPercent(fillerTTL, wrfTTL))
    mu += tbl.md() + "\n\n"   
    
    # dump plotData for use
    plotDir = f'{VISTA_DATA_BASE_DIR}{stationNo}/TmpWorking/'
    print(f'Serializing vizData to {plotDir}')
    json.dump(plotData, open(f'{plotDir}plotDataIFC.json', "w"), indent=4)
    
    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    print(f'Serializing Report to {userSiteDir}')
    open(f'{userSiteDir}ifcs.md', "w").write(mu) 
    
"""
TMP HERE - move to IFCRawDataAnalysis.py for raw data analysis of IFC pieces
... not part of .md flow per se

Want to see:
- [ ] consult report_text (vs doc)
  ... does filler get report_text off placer for its consult or its doc?
- [I] documents on Placer and Filler side

For Placer side first, dump documents of travel, telereading and common ... as
traverse st's, can get docs dumped. Different groups possible ex/ Travel
and TeleReading etc. and from Placer or Filler sides

So Far:
- for fraction of TELER with small number notes => will see a narrative ...

        "report_text": "Please refer to Inter-facility Consult for results.\n\nA
utomatically generated note - signature not required.\n \n  Electronically Filed
: {DAY}\n                    by: {author-dictator}\n
 ",

ie/ only TELER has document ref and then 

TODO:
- "reason_for_request" ... does this have initial VIRS etc text? [same remotely?] ie/ just a "text copy"?
"""
import shelve
from fmqlutils.cacher.cacherUtils import FMQLReplyStore, FilteredResultIterator
class DumpAndAnalyzeData:
    def __init__(self, stationNo, cutDate):
        self.__docsByGroupLabel = defaultdict(list)
        
        cutDateDT = datetime.strptime(cutDate.split("T")[0], "%Y-%m-%d")
        yrBeforeDateDT = cutDateDT - relativedelta(years=1)
        yrBeforeDate = datetime.strftime(yrBeforeDateDT, "%Y-%m-%d")
        onAndAfterDay = yrBeforeDate
        print(f'Calculated year before date is {yrBeforeDate}')
        
        tmpWorkingDir = f'{VISTA_DATA_BASE_DIR}{stationNo}/TmpWorking/'
        all8925ByIENDBFL = f'{tmpWorkingDir}all8925ByIEN{onAndAfterDay}.db'
        if os.path.isfile(all8925ByIENDBFL):
            print(f'8925 Types DB reduction in {all8925ByIENDBFL} done already')
            self.__8925ByIEN = shelve.open(all8925ByIENDBFL, flag='r')
            return
        dataLocn = f'{VISTA_DATA_BASE_DIR}{stationNo}/Data/'
        if not os.path.isfile(dataLocn + "8925-0.zip"):
            raise Exception("8925 not available for reduction")
        print(f'No 8925 Reduction for {onAndAfterDay} on so making')
        store = FMQLReplyStore(dataLocn)
        startAtReply = store.firstReplyFileOnOrAfterCreateDay("8925", "entry_date_time", onAndAfterDay)
        if startAtReply == "":
            raise Exception("Can't find a start at reply for {}".format("8925"))
        print("Starting with reply {}".format(startAtReply))
        resourceIter = FilteredResultIterator(dataLocn, "8925", startAtReply=startAtReply)
        all8925ByIEN = shelve.open(all8925ByIENDBFL)
        for i, resource in enumerate(resourceIter, 1):
            if i % 100000 == 0:
                print("Another 100K to {:,}".format(i))
            all8925ByIEN[str(resource["_id"].split("-")[1])] = resource
        all8925ByIEN.close()    
        self.__8925ByIEN = shelve.open(all8925ByIENDBFL, flag='r')

    """
    Ex/ dumpDocsOfST('placerTeleR', st)
    
    Only TELER has tiu_result_narrative (and associated_results)
    See limited # of TELER where doc has 'see elsewhere' text but where?
    - associated_results <=> tiu_result_narritive
    - remote_results == VPTR for 8925's remotely [ie/ doc not transferred]
    
    No Document Locally: VIRS, Travel etc - all have remote doc refs
    
    OTHER:
    - note the "from" 44's in sts ... need these for all service types or see it in patient_location
    """
    def dumpDocsOfST(self, groupLabel, st):
        if "tiu_result_narrative" not in st:
            return
        if "associated_results" not in st:
            raise Exception("Expected 'associated_results' if 'tiu_result_narrative'")
        if "byValueCount" not in st["tiu_result_narrative"]:
            return
        for ref in st["tiu_result_narrative"]["byValueCount"]:
            tiuIEN = str(re.search(r'\[8925-(\d+)\]', ref).group(1))
            self.__docsByGroupLabel[groupLabel].append(self.__8925ByIEN[tiuIEN])
            # print(json.dumps(st, indent=4))
        
    def finish(self):
        for groupLabel in self.__docsByGroupLabel:
            json.dump(self.__docsByGroupLabel[groupLabel], open(f'{groupLabel}.json', "w"), indent=4)
            print(f'Look in {groupLabel}.json for documents of {groupLabel}')
    
"""
> https://hackingandslacking.com/using-hierarchical-indexes-with-pandas-38057cf1d7e
... call multikeys a hierarchy ... order of it

[can even do on columns ie/ subtypeid and then can group cols (props) by another prop
=> of individual user if not in primary breakdown]

[pivottable needed]

Typer has multiple indexes (subtypeid) and then value ranges for all properties (ie/ not individual rows).

TODO: replicate in DF ...
- type123YR1, sts123YR1All = splitTypeDatas(stationNo, "123", reductionLabel="YR1", expectSubTypeProperties=["cprs_status", "routing_facility", "to_service"])
... comparing size and lookup effectiveness

And note, no need to flatten back out ...
- lastSeasonDF = nhlDF.xs(20182019, level='season')  # Group by last season

OTHER: encapsulate DF version in a typer class with setitem/getitem and static props
 X.prop etc. => can put behind as migrate?
"""
def typerToDFPlay(stationNo):

    """
    Walk one year 123 to DF
    """
    pass
        
# ################################# DRIVER #######################
               
# try argparse vs other cli
def main():
    """
    Specify only Station No if want .md and plot DATA made
    Specify 'PLOT' too if want to generate plots from pre-generated plot data
    """
    
    assert sys.version_info >= (3, 6)

    try:
        stationNo = sys.argv[1]
    except IndexError:
        raise SystemExit("Usage _EXE_ STATIONNO [PLOT]")

    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    if not os.path.isdir(userSiteDir):
        raise SystemExit("Expect User Site to already exist with its basic contents")
        
    if len(sys.argv) > 2:
        if sys.argv[2] == "PLOT":
            makePlots(stationNo, "IFC")
            return
        raise Exception("Only argument allowed is 'PLOT'")

    # <=> YR1 (fits with YR1E of 3.081)
    meta = metaOfVistA(stationNo)
    vistaName = "VistA" if "name" not in meta else meta["name"]
    cutDate = meta["cutDate"]
    cutDateDT = datetime.strptime(cutDate.split("T")[0], "%Y-%m-%d")
    yrBeforeDateDT = cutDateDT - relativedelta(years=1)
    yrBeforeDate = datetime.strftime(yrBeforeDateDT, "%Y-%m-%d")
    onAndAfterDay = yrBeforeDate
    upToDay = cutDate
    print("Reporting IFCs for YR1 (from {} to {})".format(onAndAfterDay, upToDay))

    webReportIFC(stationNo, meta.get("name", "VistA"), onAndAfterDay, upToDay)
        
if __name__ == "__main__":
    main()
